"""Configuration management - json interface
"""

import json
import logging
import os
import warnings

logger = logging.getLogger(__name__)


def load_spec(fname):

    with open(fname, "r+") as config_file:
        spec = json.load(config_file)
    return spec


def save_spec(spec, fname, overwrite=False):
    """Saves a specification file to json

    Parameters
    ----------

    spec : dict,
      The specification nested dictionaries to be saved
    fname: str,
      The file name, can be a path
    overwrite: bool,
      Wheter or not to overwrite an existing file, default False

    """
    if not overwrite:
        if os.path.isfile(fname):
            raise IOError(
                """%s exists and overwriting is prevented
Please set `overwrite` to True
"""
                % fname
            )
    with open(fname, "w+") as config_file:
        json.dump(spec, config_file)



import logging
from ...topology.bulk_topology import IH_transition, HI_transition
from .actions import shrink, contract

logger = logging.getLogger(__name__)


default_apoptosis_spec = {
    "cell_id": -1,
    "contract_rate": 2.0,
    "critical_area": 1e-2,
    "shrink_rate": 0.4,
    "critical_volume": 0.1,
}


def apoptosis(monolayer, manager, **kwargs):
    """Apoptotic behavior

    Parameters
    ----------
    monolayer : a :cass:`Monolayer` object
    manager : a :class:`EventManager` object

    Optional Parameters
    -------------------
    cell_id : int
       id of the apoptotic cell
    contract_rate : float, default 2.
    critical_area : float, default 1e-2,
    shrink_rate : float, default 0.4,
    critical_volume : float, default 0.1,
    """
    apoptosis_spec = default_apoptosis_spec
    apoptosis_spec.update(**kwargs)

    cell = apoptosis_spec["cell_id"]
    if cell == -1:
        return
    done = False

    cell_to_face = monolayer.get_orbits("cell", "face")
    try:
        apical_face = monolayer.face_df[
            (monolayer.face_df.index.isin(cell_to_face[cell]))
            & (monolayer.face_df.segment == "apical")
        ].index[0]
    except Exception:  # TODO fix that
        apical_face = None
    # Apical face has already been removed.
    # It needs to eliminate lateral face until obtain a cell with 4 faces.
    if apical_face is None:
        faces_in_cell = monolayer.face_df.loc[cell_to_face[cell].unique()]
        if len(faces_in_cell) > 4:
            # Remove lateral face with 3 sides
            face_to_eliminate = faces_in_cell[
                (faces_in_cell.segment == "lateral") & (faces_in_cell.num_sides == 3)
            ].index[0]

            prev_nums = {
                "edge": monolayer.Ne,
                "face": monolayer.Nf,
                "vert": monolayer.Nv,
            }
            HI_transition(monolayer, face_to_eliminate)
            monolayer.face_df.loc[prev_nums["face"] :, "contractility"] = 0
            done = False
        elif len(faces_in_cell) == 4:
            # Volume reduction
            if monolayer.cell_df.loc[cell, "vol"] > apoptosis_spec["critical_volume"]:
                shrink(monolayer, cell, apoptosis_spec["shrink_rate"])
                done = False
            else:
                done = True
    else:
        # Contract apical surface until reached a critical area
        if monolayer.face_df.loc[apical_face, "area"] > apoptosis_spec["critical_area"]:
            contract(monolayer, apical_face, apoptosis_spec["contract_rate"], True)
            done = False

        elif (
            monolayer.face_df.loc[apical_face, "area"]
            <= apoptosis_spec["critical_area"]
        ):
            # Reduce neighbours for the apical face (until 3)
            if monolayer.face_df.loc[apical_face, "num_sides"] > 3:
                e_min = monolayer.edge_df[monolayer.edge_df["face"] == apical_face][
                    "length"
                ].idxmin()
                prev_nums = {
                    "edge": monolayer.Ne,
                    "face": monolayer.Nf,
                    "vert": monolayer.Nv,
                }

                monolayer.settings["threshold_length"] = 1e-3
                IH_transition(monolayer, e_min)
                monolayer.face_df.loc[prev_nums["face"] :, "contractility"] = 0

                done = False

            elif monolayer.face_df.loc[apical_face, "num_sides"] == 3:
                prev_nums = {
                    "edge": monolayer.Ne,
                    "face": monolayer.Nf,
                    "vert": monolayer.Nv,
                }
                HI_transition(monolayer, apical_face)
                done = False

    if not done:
        manager.append(apoptosis, **apoptosis_spec)



import numpy as np
import pandas as pd

try:
    from vispy.io import write_mesh
except ImportError:
    print("You need vispy to use the .OBJ export")

import logging

logger = logging.getLogger(name=__name__)


def save_triangulated(filename, eptm):

    vertices, faces = eptm.triangular_mesh(eptm.coords, False)
    write_mesh(
        filename,
        vertices=vertices,
        faces=faces,
        normals=None,
        texcoords=None,
        overwrite=True,
    )
    logger.info("Saved %s as a trianglulated .OBJ file", eptm.identifier)


def save_junction_mesh(filename, eptm):

    vertices, faces, normals = eptm.vertex_mesh(eptm.coords, vertex_normals=True)

    write_mesh(
        filename,
        vertices=vertices,
        faces=faces,
        normals=normals,
        texcoords=None,
        overwrite=True,
        reshape_faces=False,
    )  # GH 1155
    logger.info("Saved %s as a junction mesh .OBJ file", eptm.identifier)


def write_splitted_cells(*args, **kwargs):
    logger.warning("Deprecated, use `save_splitted_cells` instead")
    save_splitted_cells(*args, **kwargs)


def save_splitted_cells(fname, sheet, epsilon=0.1):

    coords = sheet.coords
    up_srce = sheet.upcast_srce(sheet.vert_df[coords])
    up_trgt = sheet.upcast_trgt(sheet.vert_df[coords])
    up_face = sheet.upcast_face(sheet.face_df[coords])
    up_srce = (up_srce - up_face) * (1 - epsilon) + up_face
    up_trgt = (up_trgt - up_face) * (1 - epsilon) + up_face

    cell_faces = pd.concat([sheet.face_df[coords], up_srce, up_trgt], ignore_index=True)
    Ne, Nf = sheet.Ne, sheet.Nf

    triangles = np.vstack(
        [sheet.edge_df["face"], np.arange(Ne) + Nf, np.arange(Ne) + Ne + Nf]
    ).T
    write_mesh(
        fname,
        cell_faces.values,
        triangles,
        normals=None,
        texcoords=None,
        overwrite=True,
    )



import logging
import random
import numpy as np

from ...utils.decorators import cell_lookup
from ...topology.bulk_topology import IH_transition, HI_transition
from .actions import shrink, contract, relax, ab_pull, ab_pull_edge
from .basic_events import contraction
from ..sheet.basic_events import contraction as sheet_contraction
from ..sheet.delamination_events import _neighbor_contractile_increase

logger = logging.getLogger(__name__)


default_constriction_spec = {
    "cell_id": -1,
    "cell": -1,
    "contract_rate": 2.0,
    "critical_area": 1e-2,
    "shrink_rate": 0.4,
    "critical_volume": 0.1,
    "radial_tension": 1.0,
    "contract_neighbors": True,
    "critical_area_neighbors": 10,
    "contract_span": 2,
    "basal_contract_rate": 1.001,
    "current_traction": 0,
    "max_traction": 30,
    "contraction_column": "contractility",
    "with_rearrangement": False,
    "critical_area_reduction": 5,
}


@cell_lookup
def constriction(monolayer, manager, **kwargs):
    """Constriction behavior

    Parameters
    ----------
    monolayer : a :cass:`Monolayer` object
    manager : a :class:`EventManager` object
    """
    constriction_spec = default_constriction_spec
    constriction_spec.update(**kwargs)

    cell = constriction_spec["cell"]
    contract_rate = constriction_spec["contract_rate"]

    # Find the apical face id of the cell
    list_face_in_cell = monolayer.get_orbits("cell", "face")
    faces_in_cell = monolayer.face_df.loc[list_face_in_cell[cell].unique()]
    if "apical" in faces_in_cell["segment"].values:
        apical_face_id = faces_in_cell[faces_in_cell["segment"] == "apical"].index[0]
    else:
        apical_face_id = None

    if apical_face_id is None and constriction_spec["with_rearrangement"]:
        _constrict_apical(faces_in_cell, monolayer, constriction_spec)

    else:
        apical_face_area = monolayer.face_df.loc[apical_face_id, "area"]

        if apical_face_area > constriction_spec["critical_area"]:
            contract(
                monolayer,
                apical_face_id,
                contract_rate,
                True,
                constriction_spec["contraction_column"],
            )

            if (constriction_spec["contract_neighbors"]) & (
                apical_face_area < constriction_spec["critical_area_neighbors"]
            ):

                sheet = monolayer.get_sub_sheet("apical")

                neighbors = sheet.get_neighborhood(
                    apical_face_id, constriction_spec["contract_span"]
                ).dropna()
                neighbors["id"] = sheet.face_df.loc[neighbors.face, "id"].values

                # remove cell which are not mesodermal
                ectodermal_cell = sheet.face_df.loc[neighbors.face][
                    ~sheet.face_df.loc[neighbors.face, "is_mesoderm"]
                ].id.values

                neighbors = neighbors.drop(
                    neighbors[neighbors.id.isin(ectodermal_cell)].index
                )

                manager.extend(
                    [
                        (
                            sheet_contraction,
                            _neighbor_contractile_increase(neighbor, constriction_spec),
                        )
                        for _, neighbor in neighbors.iterrows()
                    ]
                )

        if not constriction_spec["with_rearrangement"]:
            # we're done
            return

        # If apical face has already been removed
        # Need to eliminate lateral face until obtain cell with 4 faces

        # If apical face area are under threshold
        if apical_face_area < constriction_spec["critical_area_reduction"]:
            # Reduce neighbours for the apical face (until 3)
            if monolayer.face_df.loc[apical_face_id, "num_sides"] > 3:
                # "Remove" the shortest edge
                e_min = monolayer.edge_df[monolayer.edge_df["face"] == apical_face_id][
                    "length"
                ].idxmin()
                prev_nums = {
                    "edge": monolayer.Ne,
                    "face": monolayer.Nf,
                    "vert": monolayer.Nv,
                }

                monolayer.settings["threshold_length"] = 1e-3
                IH_transition(monolayer, e_min)
                monolayer.face_df.loc[prev_nums["face"] :, "contractility"] = 0

            elif monolayer.face_df.loc[apical_face_id, "num_sides"] == 3:
                prev_nums = {
                    "edge": monolayer.Ne,
                    "face": monolayer.Nf,
                    "vert": monolayer.Nv,
                }
                HI_transition(monolayer, apical_face_id)

    manager.append(constriction, **constriction_spec)


def _constrict_apical(faces_in_cell, monolayer, constriction_spec):

    cell = constriction_spec["cell"]
    current_traction = constriction_spec["current_traction"]

    if len(faces_in_cell) > 4:
        # Remove lateral face with 3 sides

        three_lateral = faces_in_cell[
            (faces_in_cell.segment == "lateral") & (faces_in_cell.num_sides == 3)
        ].index
        if len(three_lateral):
            face_to_eliminate = three_lateral[0]
        prev_Nf = monolayer.Nf
        HI_transition(monolayer, face_to_eliminate)
        monolayer.face_df.loc[prev_Nf:, "contractility"] = 0

    elif len(faces_in_cell) == 4:
        if monolayer.cell_df.loc[cell, "vol"] > constriction_spec["critical_volume"]:
            shrink(monolayer, cell, constriction_spec["shrink_rate"])

        if current_traction < constriction_spec["max_traction"]:
            list_vert_in_cell = monolayer.get_orbits("cell", "srce")
            vert_in_cell = monolayer.vert_df.loc[list_vert_in_cell[4].unique()]
            apical_vert = vert_in_cell[vert_in_cell["segment"] == "apical"].index[0]
            list_vert_connected_to_apical = monolayer.edge_df[
                monolayer.edge_df["srce"] == apical_vert
            ]["trgt"].unique()
            opposite_vert = -1
            for v in list_vert_connected_to_apical:
                if v not in (vert_in_cell.index):
                    opposite_vert = v
            if opposite_vert > -1:
                edges_to_pull = monolayer.edge_df[
                    (
                        (monolayer.edge_df.trgt == apical_vert)
                        & (monolayer.edge_df.srce == opposite_vert)
                    )
                    | (
                        (monolayer.edge_df.srce == apical_vert)
                        & (monolayer.edge_df.trgt == opposite_vert)
                    )
                ].index

                ab_pull_edge(
                    monolayer, edges_to_pull, constriction_spec["radial_tension"], True
                )
                current_traction += 1
                constriction_spec.update({"current_traction": current_traction})



"""
Vertex model for an Epithelial sheet (see definitions).

Depends on the sheet vertex geometry functions.
"""
from . import effectors
from .factory import model_factory

SheetModel = model_factory(
    [
        effectors.LineTension,
        effectors.FaceContractility,
        effectors.FaceVolumeElasticity,
    ],
    effectors.FaceVolumeElasticity,
)



"""Monolayer epithelium objects
"""

import logging
import numpy as np
import pandas as pd

from scipy.spatial import cKDTree
from .objects import Epithelium
from .sheet import Sheet
from ..geometry.bulk_geometry import BulkGeometry

logger = logging.getLogger(name=__name__)


class Monolayer(Epithelium):
    """
    3D monolayer epithelium
    """

    def __init__(self, name, datasets, specs=None, coords=None):

        super().__init__(name, datasets, specs, coords)
        self.vert_df["is_active"] = 1
        self.cell_df["is_alive"] = 1
        self.face_df["is_alive"] = 1
        self.reset_topo()
        BulkGeometry.update_all(self)

    @classmethod
    def from_flat_sheet(cls, name, apical_sheet, specs, thickness=1):
        from ..generation import extrude

        datasets = extrude(
            apical_sheet.datasets, method="translation", vector=[0, 0, -thickness]
        )

        mono = cls(name, datasets, specs)
        mono.reset_topo()
        return mono

    def segment_index(self, segment, element):
        df = getattr(self, "{}_df".format(element))
        return df[df["segment"] == segment].index

    @property
    def lateral_faces(self):
        return self.segment_index("lateral", "face")

    @property
    def apical_faces(self):
        return self.segment_index("apical", "face")

    @property
    def basal_faces(self):
        return self.segment_index("basal", "face")

    @property
    def lateral_edges(self):
        return self.segment_index("lateral", "edge")

    @property
    def apical_edges(self):
        return self.segment_index("apical", "edge")

    @property
    def basal_edges(self):
        return self.segment_index("basal", "edge")

    @property
    def apical_verts(self):
        return self.segment_index("apical", "vert")

    @property
    def basal_verts(self):
        return self.segment_index("basal", "vert")

    @property
    def lateral_verts(self):
        return self.segment_index("lateral", "vert")

    def get_sub_sheet(self, segment):
        """ Returns a :class:`Sheet` object of the corresponding
        segment

        Parameters
        ----------
        segment: str, the corresponding segment, wether 'apical' or 'basal'

        """
        datasets = {
            element: self.datasets[element].loc[self.segment_index(segment, element)]
            for element in ["edge", "face", "vert"]
        }
        specs = {k: self.specs[k] for k in ["face", "edge", "vert", "settings"]}
        return Sheet(self.identifier + segment, datasets, specs)

    def guess_vert_segment(self, vert):
        """Infers the vertex segment from its surrounding edges.

        """
        v_edges = self.edge_df[self.edge_df["srce"] == vert]
        if v_edges.shape[0] == 0:
            logger.info("Vertex %d not found", vert)
            return
        if v_edges.shape[0] == 12:
            self.vert_df.loc[vert, ["segment"]] = "lateral"
            return
        intersect = {"apical", "basal"}.intersection(v_edges["segment"])
        if len(intersect) == 2:
            logger.info("Segment of vertex %d could not be determined", vert)
            self.vert_df.loc[vert, ["segment"]] = "unknown"
        elif not intersect:
            self.vert_df.loc[vert, ["segment"]] = "lateral"
        else:  # intersect is {"apical"} or {"basal"}
            self.vert_df.loc[vert, ["segment"]], = intersect

    def guess_face_segment(self, face):
        """Infers the face segment from its surrounding edges.

        """
        face_edges = self.edge_df[self.edge_df.face == face]
        if face_edges.shape[0] == 0:
            logger.info("face %d not found", face)
        v_segments = set(self.vert_df.loc[face_edges["srce"], "segment"])
        if len(v_segments) == 2:
            self.face_df.loc[face, "segment"] = "lateral"
        elif len(v_segments) == 1:
            new_segment, = v_segments
            self.face_df.loc[face, "segment"] = new_segment


class MonolayerWithLamina(Monolayer):
    """
    3D monolayer epithelium with a lamina meshing
    """

    def __init__(self, name, datasets, specs=None, coords=None):

        super().__init__(name, datasets, specs, coords)

        BulkGeometry.update_all(self)
        self.reset_index()
        from ..generation import subdivide_faces

        subdivided = subdivide_faces(self, self.basal_faces)
        for name, df in subdivided.items():
            setattr(self, "{}_df".format(name), df)
        self.reset_index()
        self.reset_topo()

        subdiv_edges = self.edge_df[self.edge_df["subdiv"] == 1].index
        self.edge_df.loc[subdiv_edges, "segment"] = "basal"

        subdiv_verts = self.vert_df[self.vert_df["subdiv"] == 1].index
        self.vert_df.loc[subdiv_verts, "segment"] = "basal"
        self.vert_df.loc[subdiv_verts, "basal_shift"] = 0.0
        self.vert_df.loc[subdiv_verts, "is_active"] = 1.0

        subdiv_verts = self.vert_df[self.vert_df["subdiv"] == 1].index
        focal_adhesions = self.vert_df.loc[subdiv_verts]

        max_dist = self.edge_df.length.dropna().median() * 1.74
        lamina_tree = cKDTree(focal_adhesions[self.coords].values)
        lamina_edges = pd.DataFrame(
            [[i, j] for i, j in lamina_tree.query_pairs(max_dist, eps=1e-3)],
            columns=["srce", "trgt"],
        )
        lamina_edges.index.name = "edge"
        lamina_edges["srce"] = focal_adhesions.index[lamina_edges["srce"]]
        lamina_edges["trgt"] = focal_adhesions.index[lamina_edges["trgt"]]
        # place holder face and cell
        lamina_face = self.face_df.index.max() + 1
        lamina_edges["face"] = lamina_face
        self.face_df.append(self.face_df.iloc[0].copy())
        self.face_df.loc[lamina_face, "is_alive"] = 0

        lamina_cell = self.cell_df.index.max() + 1
        lamina_edges["cell"] = lamina_cell
        self.cell_df.append(self.cell_df.iloc[0].copy())
        self.cell_df.loc[lamina_cell, "is_alive"] = 0

        lamina_edges.index += self.edge_df.index.max() + 1
        lamina_edges["segment"] = "lamina"
        lamina_edges["subdiv"] = 0
        self.edge_df = pd.concat([self.edge_df, lamina_edges], sort=True)
        self.reset_topo()
        BulkGeometry.update_all(self)

    @property
    def lamina_edges(self):
        return self.segment_index("lamina", "edge")



"""
Core definitions
"""
import warnings
import logging
from collections import deque
import numpy as np
import pandas as pd
from copy import deepcopy
from ..utils.utils import set_data_columns, spec_updater

log = logging.getLogger(name=__name__)


class Epithelium:
    """Base class defining a connective tissue in 2D or 3D.
    """

    def __init__(self, identifier, datasets, specs=None, coords=None, maxbackup=5):
        """Creates an epithelium

        Parameters
        ----------
        identifier : string
        datasets : dictionary of dataframes
            The keys correspond to the different geometrical elements
            constituting the epithelium:

            * `vert` contains a dataframe of vertices,
            * `edge` contains a dataframe of *oriented* half-edges between vertices,
            * `face` contains a dataframe of polygonal faces enclosed by half-edges,
            * `cell` contains a dataframe of polyhedral cells delimited by faces,

        specs : nested dictionnary of specifications
            The first key designs the element name: (`vert`, `edge`, `face`, `cell`),
            corresponding to the respective dataframes attribute in the dataset.
            The second level keys design column names of the above dataframes.
            For exemple:
            .. code::
                specs = {
                    "face": {
                        ## Face Geometry
                        "perimeter": 0.,
                        "area": 0.,
                        ## Coordinates
                        "x": 0.,
                        "y": 0.,
                        "z": 0.,
                        ## Topology
                        "num_sides": 6,
                        ## Masks
                        "is_alive": True},
                    "vert": {
                        ## Coordinates
                        "x": 0.,
                        "y": 0.,
                        "z": 0.,
                        ## Masks
                        "is_active": True},
                    "edge": {
                        ## Connections
                        "srce": 0,
                        "trgt": 1,
                        "face": 0,
                        "cell": 0,
                        ## Coordinates
                        "dx": 0.,
                        "dy": 0.,
                        "dz": 0.,
                        "length": 0.,
                        ## Normals
                        "nx": 0.,
                        "ny": 0.,
                        "nz": 1.}
                    "settings":
                        ## Custom values
                        "geometry": "flat"
                    }

        Note
        ----
        For efficiency reasons, we have to maintain a monotonous RangeIndex
        for each dataset. Thus, **the index of an element can change**,
        and should not be used as an identifier.

        """
        # backup container
        # TODO: pass the max backup number as a config argument
        self._backups = deque(maxlen=maxbackup)

        # each of those has a separate dataframe, as well as entries in
        # the specification files
        _frame_types = {"edge", "vert", "face", "cell"}
        self.identifier = identifier
        if not set(datasets).issubset(_frame_types):
            raise ValueError(
                "The `datasets` dictionnary should"
                " contain keys in {}".format(_frame_types)
            )
        self.datasets = datasets
        self.data_names = list(datasets.keys())
        self.element_names = ["srce", "trgt", "face", "cell"][: len(self.data_names)]
        # Infer specs from the first rows of the datasets
        if specs is None:
            specs = {elem: df.iloc[0].to_dict() for elem, df in datasets.items()}
        if "settings" not in specs:
            specs["settings"] = {}

        self.specs = specs
        self.update_specs(specs, reset=False)

        if coords is None:
            coords = [c for c in "xyz" if c in datasets["vert"].columns]

        self.coords = coords
        # edge's dx, dy, dz
        self.dcoords = ["d" + c for c in self.coords]
        # edge's unit length vector
        self.ucoords = ["u" + c for c in self.coords]

        self.dim = len(self.coords)
        # edge's normals
        if self.dim == 3:
            self.ncoords = ["n" + c for c in self.coords]

        self._bad = None
        self.bbox = None
        if "is_active" in self.vert_df.columns:
            self.active_verts = self.vert_df[self.vert_df.is_active == 1].index
        else:
            self.active_verts = self.vert_df.index
        self.set_bbox()

        self.position_buffer = None
        self.topo_changed = False

    @property
    def vert_df(self):
        """The face :class:`pd.DataFrame` containing vertex associated
        data e.g. the position of each vertex.
        """
        return self.datasets["vert"]

    @vert_df.setter
    def vert_df(self, value):
        self.datasets["vert"] = value

    @property
    def face_df(self):
        """The face :class:`pd.DataFrame` containing face associated
        data e.g. the position of their center or their area
        """
        return self.datasets["face"]

    @face_df.setter
    def face_df(self, value):
        self.datasets["face"] = value

    @property
    def edge_df(self):
        """The edge :class:`pd.DataFrame` containing edge associated
        data e.g. their length.This dataframe also contains the whole
        connexion of the epithelium through the `"srce", "trgt", "face", "cell"`
        indices. In 2D, a half-edge is associated with a single (face, srce, trgt)
        positively oriented triangle. In 3D, the (cell, face, srce, trgt)
        positively oriented terahedron is also unique.
        """
        return self.datasets["edge"]

    @edge_df.setter
    def edge_df(self, value):
        self.datasets["edge"] = value

    @property
    def cell_df(self):
        """The cell :class:`pd.DataFrame` containing cell associated
        data e.g. the position of their center or their volume
        """
        return self.datasets.get("cell", None)

    @cell_df.setter
    def cell_df(self, value):
        self.datasets["cell"] = value

    def copy(self, deep_copy=True):
        """
        Returns a copy of the epithelium

        Parameters
        ----------
        deep_copy : bool, default True
            if True, use a copy of the original object's datasets
            to create the new object. If False, datasets are not copied
        """
        if deep_copy:
            datasets = {element: df.copy() for element, df in self.datasets.items()}
            specs = deepcopy(self.specs)

        else:  # pragma: no cover
            log.info("New epithelium object from %s without deep copy", self.identifier)
            datasets = self.datasets
            specs = self.specs

        identifier = self.identifier + "_copy"

        new = type(self)(identifier, datasets, specs=specs, coords=self.coords)

        return new

    def backup(self):
        """Creates a copy of self and keeps a reference to it
        in the self._backups deque.

        """
        log.info("Backing up")
        self._backups.append(self.copy(deep_copy=True))

    def restore(self):
        """Resets the eptithelium data to its last backed up state

        A copy of the current state prior to restoring is kept in the
        `_bad` attribute for inspection.
        Calling this method multiple times (without calling backup) will
        go back in the epithelium backups.
        """

        log.info("Restoring")
        log.info(
            "a copy of the unrestored epithelium" " is stored in the `_bad` attribute"
        )
        bck = self._backups.pop()
        self._bad = self.copy(deep_copy=True)
        self.datasets = bck.datasets
        self.specs = bck.specs

    @property
    def settings(self):
        """ Accesses the `specs['settings']` dictionnary
        """
        return self.specs["settings"]

    def update_specs(self, new, reset=False):
        """Recursively updates the `self.specs` nested dictionnary,
        and set the new values to the corresponding columns
        in the datasets. If reset is `True`, existing values
        will be overwritten.
        """
        spec_updater(self.specs, new)
        set_data_columns(self.datasets, new, reset)

    def update_num_sides(self):
        """Updates the number of half-edges around the faces.
        The data is registered in the `"num_sides"` column of
        `self.face_df`.
        """
        self.face_df["num_sides"] = self.edge_df.face.value_counts()

    def update_num_faces(self):
        """Updates the number of faces around the cells.
        The data is registered in the `"num_faces"` column of
        `self.cell_df`.
        """
        self.cell_df["num_faces"] = self.edge_df.groupby("cell").apply(
            lambda df: df["face"].unique().size
        )
        self.cell_df["num_ridges"] = self.edge_df.cell.value_counts()

    def reset_topo(self):
        """Recomputes the number of sides for the faces and the
        number of faces for the cells.
        """
        log.debug("Resetting topology")
        self.update_num_sides()
        if "is_active" in self.vert_df.columns:
            self.active_verts = self.vert_df[self.vert_df.is_active == 1].index
        if "cell" in self.data_names:
            self.update_num_faces()

    @property
    def Nv(self):
        """The number of vertices in the epithelium
        """
        return self.vert_df.shape[0]

    @property
    def Ne(self):
        """The number of edges in the epithelium
        """
        return self.edge_df.shape[0]

    @property
    def Nf(self):
        """The number of faces in the epithelium
        """
        return self.face_df.shape[0]

    @property
    def Nc(self):
        """The number of cells in the epithelium
        """
        if "cell" in self.data_names:
            return self.cell_df.shape[0]
        elif "face" in self.data_names:
            return self.face_df.shape[0]
        return None

    def _upcast(self, idx, df):

        ## Assumes a flat index
        upcast = df.take(idx)
        try:
            upcast.index = self.edge_df.index
        except AttributeError:
            if len(upcast.shape) == 1:
                upcast = pd.Series(upcast, index=self.edge_df.index)
            else:
                upcast = pd.DataFrame(upcast, index=self.edge_df.index)
        return upcast

    def upcast_cols(self, element, columns):
        """Syntactic sugar to upcast from the epithelium datasets.

        Parameters
        ----------
        element: {'srce'|'trgt'|'face'|'cell'}
           corresponding self.edge_df column over which to index
           if element is 'srce' or 'trgt', the upcast data will be
           taken form self.vert_df
        columns: index
           the column(s) to be taken from the input dataset.

        """
        if element in ["srce", "trgt"]:
            dataset = "vert"
        else:
            dataset = element
        return self._upcast(self.edge_df[element], self.datasets[dataset][columns])

    def upcast_srce(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each source entry.

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nv` elements.
          If a string is passed it should be a column of `self.vert_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["srce"]`
        """
        if isinstance(df, str):
            df = self.vert_df[df]
        return self._upcast(self.edge_df["srce"], df)

    def upcast_trgt(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each target entry

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nv` elements.
          If a string is passed it should be a column of `self.vert_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["trgt"]`
        """
        if isinstance(df, str):
            df = self.vert_df[df]
        return self._upcast(self.edge_df["trgt"], df)

    def upcast_face(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each face entry

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nf` elements.
          If a string is passed it should be a column of `self.face_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["face"]`

        """
        if isinstance(df, str):
            df = self.face_df[df]
        return self._upcast(self.edge_df["face"], df)

    def upcast_cell(self, df):
        """ Reindexes input data to self.edge_df.index
        by repeating the values for each cell entry

        Parameters
        ----------
        df : :class:`pd.DataFrame`, :class:`pd.Series` :class:`np.ndarray` or string
          The data to be upcasted. If array like, should have `self.Nc` elements.
          If a string is passed it should be a column of `self.cell_df`

        Returns
        -------
        upcast_df : :class:`pd.DataFrame` or :class:`pd.Series`
          The value repeated like the values of `self.edge_df["cell"]`
        """
        if isinstance(df, str):
            df = self.cell_df[df]
        return self._upcast(self.edge_df["cell"], df)

    def _lvl_sum(self, df, lvl):
        df_ = df
        if isinstance(df, np.ndarray):
            df_ = pd.DataFrame(df, index=self.edge_df.index)
        elif isinstance(df, pd.Series):
            df_ = df.to_frame()
        elif lvl not in df.columns:
            df_ = df.copy()
        df_[lvl] = self.edge_df[lvl]
        return df_.groupby(lvl).sum()

    def sum_srce(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["srce"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "srce")

    def sum_trgt(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["trgt"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "trgt")

    def sum_face(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["face"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "face")

    def sum_cell(self, df):
        """Sums the values of the edge-indexed dataframe `df` grouped by
        the values of `self.edge_df["cell"]`

        Returns
        -------
        summed : :class:`pd.DataFrame` the summed data, indexed by the source vertices.
        """
        return self._lvl_sum(df, "cell")

    def get_orbits(self, center, periph):
        """Returns a dataframe with a `(center, edge)` MultiIndex with `periph`
        elements.

        Parameters
        ----------
        center : str,
            the name of the center element for example 'face', 'srce'
        periph : str,
            the name of the periphery elements, for example 'trgt', 'cell'

        Example
        -------
        >>> cell_verts = sheet.get_orbits('face', 'srce')
        >>> cell_verts.loc[45]
        edge
        218    75
        219    78
        220    76
        221    81
        222    90
        223    87
        Name: srce, dtype: int64

        """
        orbits = self.edge_df.groupby(center).apply(lambda df: df[periph])
        return orbits

    def idx_lookup(self, elem_id, element):
        """returns the current index of the element with the `"id"` column equal to `elem_id`

        Parameters
        ----------
        elem_id : int
          id of the element to retrieve
        element : {"vert"|"edge"|"face"|"cell"}
          the corresponding dataset.
        """
        df = self.datasets[element]["id"]
        idx = df[df == elem_id].index
        if len(idx):
            return idx[0]
        else:
            return None

    def get_neighbors(self, elem_id, elem="cell"):
        """Returns the indexes of the adjacent elements (cells or faces) of
        the element of index `elem_id`.

        Parameters
        ----------
        elem_id : int
            the index of the central element (a face or a cell)
        element : {'cell' | 'face'}, default 'cell'

        Returns
        -------
        neghbors : set
            the cells (or faces) sharing an edge with the central cell (face)
        """

        topo = self.edge_df[["srce", "trgt", elem]]
        edges = self.edge_df[self.edge_df[elem] == elem_id][["srce", "trgt"]]

        neighbors = set(
            topo[
                (topo["srce"].isin(edges["srce"]) & topo["trgt"].isin(edges["trgt"]))
                | (topo["srce"].isin(edges["trgt"]) & topo["trgt"].isin(edges["srce"]))
            ][elem]
        )
        return neighbors - {elem_id}

    def get_neighborhood(self, elem_id, order, elem="cell"):
        """Returns `elem_id` neighborhood up to a degree of `order`

        For example, if `order` is 2, it wil return the adjacent cells (or faces)
        and theses cells neighbors.

        Returns
        -------
        neighbors : pd.DataFrame with two colums, the index
            of the neighboring cell (face), and it's neighboring order

        """

        neighbors = pd.DataFrame.from_dict({elem: [elem_id], "order": [0]})

        for k in range(order + 1):
            for neigh in neighbors[neighbors["order"] == k - 1][elem]:
                new_neighs = self.get_neighbors(neigh)
                new_neighs = set(new_neighs).difference(neighbors[elem])
                orders = np.ones(len(new_neighs), dtype=np.int) * (k)
                new_neighs = pd.DataFrame.from_dict(
                    {elem: list(new_neighs), "order": orders}, dtype=np.int
                )
                neighbors = pd.concat([neighbors, new_neighs])
        return neighbors.reset_index(drop=True).loc[1:]

    def face_polygons(self, coords=None):
        """ Returns a pd.Series of arrays with the coordinates the face polygons

        Each element of the Series is a (num_sides, num_dims) array of points
        ordered counterclockwise.
        """
        if coords is None:
            coords = self.coords

        def _get_verts_pos(face):
            edges = _ordered_edges(face)
            return np.array([self.vert_df.loc[idx[0], coords] for idx in edges])

        polys = self.edge_df.groupby("face").apply(_get_verts_pos).dropna()
        return polys

    def validate(self):
        """returns True if the mesh is validated

        e.g. has only closed polygons and polyhedra
        """
        return np.alltrue(self.get_valid())

    def get_valid(self):
        """Set the 'is_valid' column to true if the faces are all closed polygons,
        and the cells closed polyhedra.
        """
        is_valid_face = self.edge_df.groupby("face").apply(_test_valid)
        is_valid = self.upcast_face(is_valid_face)
        if "cell" in self.data_names:
            is_valid_cell = self.edge_df.groupby("cell").apply(_is_closed_cell)
            is_valid = np.logical_and(is_valid, self.upcast_cell(is_valid_cell))
        self.edge_df["is_valid"] = is_valid
        return is_valid

    def get_invalid(self):
        """Returns a mask over self.edge_df for invalid faces
        """
        is_valid = self.get_valid()
        return ~is_valid

    def sanitize(self, trim_borders=False):
        """Removes invalid faces and associated vertices

        If trim_borders is True (defaults to False), there will be a single
        border edge per border face.
        """
        invalid_edges = self.get_invalid()
        if not len(invalid_edges) and trim_borders:
            from ..topology.base_topology import merge_border_edges

            merge_border_edges(self)
        self.remove(invalid_edges, trim_borders)

    def remove(self, edge_out, trim_borders=False):
        """Remove the edges indexed by `edge_out` associated with all
        the cells and faces containing those edges

        If trim_borders is True (defaults to False), there will be a single
        border edge per border face.
        """
        top_level = self.element_names[-1]
        log.info("Removing cells at the %s level", top_level)
        fto_rm = self.edge_df.loc[edge_out, top_level].unique()
        if not fto_rm.shape[0]:
            log.info("Nothing to remove")
            return
        if fto_rm.shape[0] == self.datasets[top_level].shape[0]:
            raise ValueError("sanitize would delete the whole epithlium")

        fto_rm.sort()
        log.info("%d %s level elements will be removed", len(fto_rm), top_level)

        edge_df_ = (
            self.edge_df.set_index(top_level, append=True).swaplevel(0, 1).sort_index()
        )
        to_rm = np.concatenate([edge_df_.loc[c].index.values for c in fto_rm])
        to_rm.sort()
        self.edge_df = self.edge_df.drop(to_rm)

        remaining_verts = np.unique(self.edge_df[["srce", "trgt"]])
        self.vert_df = self.vert_df.loc[remaining_verts]
        if top_level == "face":
            self.face_df = self.face_df.drop(fto_rm)
        elif top_level == "cell":
            remaining_faces = np.unique(self.edge_df["face"])
            self.face_df = self.face_df.loc[remaining_faces]
            self.cell_df = self.cell_df.drop(fto_rm)
        self.reset_index()
        self.reset_topo()
        if trim_borders:
            from ..topology.base_topology import merge_border_edges

            merge_border_edges(self)

    def cut_out(self, bbox, coords=None):
        """Returns the index of edges with at least one vertex outside of the bounding box

        Parameters
        ----------
        bbox : sequence of shape (dim, 2)
             the bounding box as (min, max) pairs for each coordinates.
        coords : list of str of len dim, default None
             the coords corresponding to the bbox.
        """
        if coords is None:
            coords = self.coords
        outs = pd.DataFrame(index=self.edge_df.index, columns=coords)
        for c, bounds in zip(coords, bbox):
            out_vert_ = (self.vert_df[c] < bounds[0]) | (self.vert_df[c] > bounds[1])
            outs[c] = self.upcast_srce(out_vert_) | self.upcast_trgt(out_vert_)

        edge_out = outs.sum(axis=1).astype(np.bool)
        return self.edge_df[edge_out].index

    def set_bbox(self, margin=0.0):
        """Sets the attribute `bbox` with pairs of values bellow
        and above the min and max of the vert coords, with a margin.
        """
        self.bbox = np.array(
            [
                [self.vert_df[c].min() - margin, self.vert_df[c].max() + margin]
                for c in self.coords
            ]
        )

    def reset_index(self):
        """Resets the datasets  to have continuous indices
        """
        log.debug("reseting index for %s", self.identifier)
        self.topo_changed = True

        # remove disconnected vertices and faces
        self.vert_df = self.vert_df.reindex(
            set(self.edge_df.srce).union(self.edge_df.trgt)
        )
        self.face_df = self.face_df.reindex(set(self.edge_df.face))

        new_vertidx = pd.Series(
            np.arange(self.vert_df.shape[0]), index=self.vert_df.index
        )
        # Here we use loc and not the take from upcast

        self.edge_df["srce"] = new_vertidx.reindex(self.edge_df["srce"]).values
        self.edge_df["trgt"] = new_vertidx.reindex(self.edge_df["trgt"]).values

        new_fidx = pd.Series(np.arange(self.face_df.shape[0]), index=self.face_df.index)

        self.edge_df["face"] = new_fidx.loc[self.edge_df["face"]].values

        self.vert_df.reset_index(drop=True, inplace=True)
        self.vert_df.index.name = "vert"

        self.face_df.reset_index(drop=True, inplace=True)
        self.face_df.index.name = "face"

        if "cell" in self.data_names:
            new_cidx = pd.Series(
                np.arange(self.cell_df.shape[0]), index=self.cell_df.index
            )
            self.edge_df["cell"] = new_cidx.loc[self.edge_df["cell"]].values
            self.cell_df.reset_index(drop=True, inplace=True)
            self.cell_df.index.name = "cell"

        self.edge_df.reset_index(drop=True, inplace=True)
        self.edge_df.index.name = "edge"

    def triangular_mesh(self, coords=None, return_mask=True):
        """
        Return a triangulation of an epithelial sheet (2D in a 3D space),
        with added edges between face barycenters and junction vertices.

        Parameters
        ----------
        coords : list of str:
          pair of coordinates corresponding to column names
          for self.face_df and self.vert_df

        return_mask : bool, optional, default True
          if True, returns `face_mask`

        Returns
        -------
        vertices : (self.Nf+self.Nv, 3) ndarray
           all the vertices' coordinates
        triangles : (self.Ne, 3) ndarray of ints
           triple of the vertices' indexes forming
           the triangular elements. For each junction edge, this is simply
           the index (srce, trgt, face). This is correctly oriented.
        face_mask: (self.Nf + self.Nv,) mask with 1 iff the vertex corresponds
           to a face center
        """
        if coords is None:
            coords = self.coords

        vertices = np.concatenate((self.face_df[coords], self.vert_df[coords]), axis=0)

        # edge indices as (Nf + Nv) * 3 array
        triangles = self.edge_df[["srce", "trgt", "face"]].values
        # The src, trgt, face triangle is correctly oriented
        # both vert_idx cols are shifted by Nf
        triangles[:, :2] += self.Nf

        # Ensure returned arrays are C-contiguous
        if not vertices.data.c_contiguous:
            vertices = np.array(vertices, order="C")

        if not triangles.data.c_contiguous:
            triangles = np.array(triangles, order="C")

        if not return_mask:
            return vertices, triangles

        warnings.warn(
            "In version 0.3, the `return_face_mask argument` will be set to False by default"
        )
        face_mask = np.arange(self.Nf + self.Nv) < self.Nf
        return vertices, triangles, face_mask

    def vertex_mesh(self, coords, vertex_normals=True):
        """ Returns the vertex coordinates and a list of vertex indices
        for each face of the tissue.
        If `vertex_normals` is True, also returns the normals of each vertex
        (set as the average of the vertex' edges), suitable for .OBJ export
        """
        vertices = self.vert_df[coords]
        faces = self.edge_df.groupby("face").apply(_ordered_vert_idxs)
        faces = faces.dropna()

        if vertex_normals:
            normals = (
                self.edge_df.groupby("srce")[self.ncoords].mean()
                + self.edge_df.groupby("trgt")[self.ncoords].mean()
            ) / 2.0
            return vertices.values, faces.values, normals.values
        return vertices.values, faces.values

    def validate_closed_cells(self):
        is_closed = self.edge_df.groupby("cell").apply(_is_closed_cell)
        return is_closed

    def get_opposite_faces(self):
        """Populates the 'opposite' column of self.face_df with the index of
        the opposite face or -1 if the face has no opposite.

        """
        face_v = self.edge_df.groupby("face").apply(lambda df: frozenset(df["srce"]))
        face_v2 = pd.Series(data=face_v.index, index=face_v.values)
        grouped = face_v2.groupby(level=0)
        cardinal = grouped.apply(len)
        if cardinal.max() > 2:
            raise ValueError(
                "Invalid topology," " incorrect faces: {}".format(cardinal.argmax())
            )
        self.face_df["opposite"] = -1

        face_pairs = []
        grouped.apply(
            lambda s: face_pairs.append(list(s.values)) if len(s) == 2 else np.nan
        ).dropna()
        face_pairs = np.array(face_pairs)
        if not face_pairs.shape[0]:
            return
        self.face_df.loc[face_pairs[:, 0], "opposite"] = face_pairs[:, 1]
        self.face_df.loc[face_pairs[:, 1], "opposite"] = face_pairs[:, 0]


def get_opposite_faces(eptm):
    warnings.warn("Deprecated, use `eptm.get_opposite_faces()` instead")
    eptm.get_opposite_faces()


def _ordered_edges(face_edges):
    """Returns "srce", "trgt" and "face" indices
    organized clockwise for each face.

    Parameters
    ----------
    face_edges: `pd.DataFrame`
        exerpt of an edge_df for a single face

    Returns
    -------
    edges: list of 3 ints
        srce, trgt, face indices, ordered
    """

    srces, trgts, faces = (face_edges[col].values for col in ["srce", "trgt", "face"])
    srce, trgt, face_ = srces[0], trgts[0], faces[0]
    edges = [[srce, trgt, face_]]
    for face_ in faces[1:]:
        srce, trgt = trgt, trgts[srces == trgt][0]
        edges.append([srce, trgt, face_])
    return edges


def _ordered_vert_idxs(face):
    try:
        return [idxs[0] for idxs in _ordered_edges(face)]
    except IndexError:
        return np.nan


def get_next_edges(sheet):
    """
    returns a pd.Series with the index of the next
    edge for each edge
    """
    next_e = sheet.edge_df.groupby("face").apply(_next_edge)
    next_e.index = next_e.index.droplevel("face")
    return next_e.sort_index()


def get_prev_edges(sheet):
    """
    returns a pd.Series with the index of the next
    edge for each edge
    """
    prev_e = sheet.edge_df.groupby("face").apply(_prev_edge)
    prev_e.index = prev_e.index.droplevel("face")
    return prev_e.sort_index()


def _next_edge(edf):

    edf["edge"] = edf.index
    next_edge = edf.set_index("srce", append=False).loc[edf["trgt"], "edge"].values
    return pd.Series(index=edf.index, data=next_edge)


def _prev_edge(edf):

    edf["edge"] = edf.index
    next_edge = edf.set_index("trgt", append=False).loc[edf["srce"], "edge"].values
    return pd.Series(index=edf.index, data=next_edge)


def _is_closed_cell(e_df):
    edges = e_df[["srce", "trgt"]]
    for _, (srce, trgt) in edges.iterrows():
        if edges[(edges["srce"] == trgt) & (edges["trgt"] == srce)].index.size != 1:
            return False
    return True


def _test_invalid(face):
    """ Returns True if the source and target sets of the faces polygon
    are different or if the face polygon is not closed
    """

    s1 = set(face["srce"])
    s2 = set(face["trgt"])
    if s1 != s2:
        return True
    ordered = np.array(_ordered_edges(face))
    if not np.all(ordered[:, 0] == np.roll(ordered[:, 1], 1)):
        return True
    return False


def _test_valid(face):
    """ Returns true iff all sources are also targets for the faces polygon
    """
    return np.logical_not(_test_invalid(face))



import pandas as pd
import numpy as np


def write_storm_csv(
    filename, points, coords=["x", "y", "z"], split_by=None, **csv_args
):
    """
    Saves a point cloud array in the storm format
    """
    columns = ["frame", "x [nm]", "y [nm]", "z [nm]", "uncertainty_xy", "uncertainty_z"]
    points = points.dropna()
    storm_points = pd.DataFrame(np.zeros((points.shape[0], 6)), columns=columns)
    storm_points[["x [nm]", "y [nm]", "z [nm]"]] = points[coords].values
    storm_points["frame"] = 1
    storm_points[["uncertainty_xy", "uncertainty_z"]] = 2.1
    # tab separated values are faster and more portable than excel
    if split_by is None:
        if not filename.endswith(".csv"):
            filename = filename + ".csv"
        storm_points.to_csv(filename, **csv_args)
    elif split_by in points.columns():
        storm_points[split_by] = points[split_by]
        # separated files by the column split_by
        storm_points.groupby(split_by).apply(
            lambda df: df.to_csv(
                "{}_{}.csv".format(filename, df[split_by].iloc[0]), **csv_args
            )
        )



"""Monolayer related behaviors

For now only apoptosis is defined, as a sequence of actions leading to the cell
disapearing from a monolayer
"""

from .apoptosis_events import apoptosis
from .delamination_events import constriction
from .basic_events import contraction
from .actions import grow, shrink, contract, contract_apical_face, ab_pull



import math
import warnings
import numpy as np
import pandas as pd
from scipy.spatial import Voronoi
from scipy import interpolate

from .. import config
from ..core.sheet import Sheet, get_outer_sheet
from ..core.objects import get_prev_edges
from ..core.objects import Epithelium
from ..topology import type1_transition
from .from_voronoi import from_3d_voronoi
from ..geometry.bulk_geometry import BulkGeometry
from ..geometry.sheet_geometry import EllipsoidGeometry, SheetGeometry

from ..utils import single_cell


class AnnularSheet(Sheet):
    """2D annular model of a cylinder-like monolayer.

    Provides syntactic sugar to access the apical, basal and
    lateral segments of the epithlium
    """

    def segment_index(self, segment, element):
        df = getattr(self, "{}_df".format(element))
        return df[df["segment"] == segment].index

    @property
    def lateral_edges(self):
        return self.segment_index("lateral", "edge")

    @property
    def apical_edges(self):
        return self.segment_index("apical", "edge")

    @property
    def basal_edges(self):
        return self.segment_index("basal", "edge")

    @property
    def apical_verts(self):
        return self.segment_index("apical", "vert")

    @property
    def basal_verts(self):
        return self.segment_index("basal", "vert")

    def reset_topo(self):
        super().reset_topo()
        self.edge_df["prev"] = get_prev_edges(self)


def generate_ring(Nf, R_in, R_out, R_vit=None, apical="in"):
    """ Generates a 2D tyssue object aranged in a ring of Nf tetragonal cells
    with inner diameter R_in and outer diameter R_out

    Parameters
    ----------
    Nf : int
        The number of cells in the tissue
    R_in : float
        The inner ring diameter
    R_out : float
        The outer ring diameter
    R_vit : float
        The vitelline membrane diameter
        (a non strechable membrane around the annulus)
    apical : str {'in' | 'out'}
        The side of the apical surface
        if "in", the apical surface is inside the annulus, facing
        the lumen as in an organoid; if 'out': the apical side is facing the
        exterior of the tissue, as in an embryo
    Returns
    -------
    eptm : :class:`AnnularSheet`
        2D annular tissue. The `R_in` and `R_out` parameters
        are stored in the class `settings` attribute.
    """
    specs = config.geometry.planar_spec()
    specs["settings"] = specs.get("settings", {})
    specs["settings"]["R_in"] = R_in
    specs["settings"]["R_out"] = R_out
    specs["settings"]["R_vit"] = R_vit

    Ne = Nf * 4
    Nv = Nf * 2
    vert_df = pd.DataFrame(
        index=pd.Index(range(Nv), name="vert"),
        columns=specs["vert"].keys(),
        dtype=float,
    )
    edge_df = pd.DataFrame(
        index=pd.Index(range(Ne), name="edge"),
        columns=specs["edge"].keys(),
        dtype=float,
    )
    face_df = pd.DataFrame(
        index=pd.Index(range(Nf), name="face"),
        columns=specs["face"].keys(),
        dtype=float,
    )

    inner_edges = np.array(
        [
            [f0, v0, v1]
            for f0, v0, v1 in zip(range(Nf), range(Nf), np.roll(range(Nf), -1))
        ]
    )

    outer_edges = np.zeros_like(inner_edges)
    outer_edges[:, 0] = inner_edges[:, 0]
    outer_edges[:, 1] = inner_edges[:, 2] + Nf
    outer_edges[:, 2] = inner_edges[:, 1] + Nf

    left_spokes = np.zeros_like(inner_edges)
    left_spokes[:, 0] = inner_edges[:, 0]
    left_spokes[:, 1] = outer_edges[:, 2]
    left_spokes[:, 2] = inner_edges[:, 1]

    right_spokes = np.zeros_like(inner_edges)
    right_spokes[:, 0] = inner_edges[:, 0]
    right_spokes[:, 1] = inner_edges[:, 2]
    right_spokes[:, 2] = outer_edges[:, 1]

    edges = np.concatenate([inner_edges, outer_edges, left_spokes, right_spokes])

    edge_df[["face", "srce", "trgt"]] = edges
    edge_df[["face", "srce", "trgt"]] = edge_df[["face", "srce", "trgt"]].astype(int)

    thetas = np.linspace(0, 2 * np.pi, Nf, endpoint=False)
    thetas += thetas[1] / 2

    thetas = thetas[::-1]
    # Setting vertices position (turning clockwise for correct orientation)
    vert_df.loc[range(Nf), "x"] = R_in * np.cos(thetas)
    vert_df.loc[range(Nf), "y"] = R_in * np.sin(thetas)
    vert_df.loc[range(Nf, 2 * Nf), "x"] = R_out * np.cos(thetas)
    vert_df.loc[range(Nf, 2 * Nf), "y"] = R_out * np.sin(thetas)

    vert_df["segment"] = "basal"
    edge_df["segment"] = "basal"
    if apical == "out":
        edge_df.loc[range(Nf, 2 * Nf), "segment"] = "apical"
        vert_df.loc[range(Nf, 2 * Nf), "segment"] = "apical"
    elif apical == "in":
        edge_df.loc[range(Nf), "segment"] = "apical"
        vert_df.loc[range(Nf), "segment"] = "apical"
    else:
        raise ValueError(
            f"apical argument not understood,"
            'should be either "in" or "out", got {apical}'
        )
    edge_df.loc[range(2 * Nf, 4 * Nf), "segment"] = "lateral"

    datasets = {"vert": vert_df, "edge": edge_df, "face": face_df}
    ring = AnnularSheet("ring", datasets, specs, coords=["x", "y"])

    ring.reset_topo()
    return ring


"""
Ellipsoid 2.5D sheet
"""


def ellipse_rho(theta, a, b):
    return ((a * math.sin(theta)) ** 2 + (b * math.cos(theta)) ** 2) ** 0.5


def get_ellipsoid_centers(a, b, c, n_zs, pos_err=0.0, phase_err=0.0):
    """
    Creates hexagonaly organized points on the surface of an ellipsoid

    Parameters
    ----------
    a, b, c: float
      ellipsoid radii along the x, y and z axes, respectively
      i.e the ellipsoid boounding box will be
      `[[-a, a], [-b, b], [-c, c]]`
    n_zs :  float
      number of cells on the z axis, typical
    pos_err : float, default 0.
      normaly distributed noise of std. dev. pos_err is added
      to the centers positions
    phase_err : float, default 0.
      normaly distributed noise of std. dev. phase_err is added
      to the centers angle ϕ

    """
    if b != a:
        warnings.warn(
            "Different half axes length along x and y"
            " axes are not supported at the moment"
        )

    dist = c / n_zs
    theta = -np.pi / 2
    thetas = []
    while theta < np.pi / 2:
        theta = theta + dist / ellipse_rho(theta, a, c)
        thetas.append(theta)

    thetas = np.array(thetas).clip(-np.pi / 2, np.pi / 2)
    zs = c * np.sin(thetas)

    # np.linspace(-c, c, n_zs, endpoint=False)
    # thetas = np.arcsin(zs/c)
    av_rhos = (a + b) * np.cos(thetas) / 2
    n_cells = np.ceil(av_rhos / dist).astype(np.int)

    phis = np.concatenate(
        [
            np.linspace(-np.pi, np.pi, nc, endpoint=False) + (np.pi / nc) * (i % 2)
            for i, nc in enumerate(n_cells)
        ]
    )

    if phase_err > 0:
        phis += np.random.normal(scale=phase_err * np.pi, size=phis.shape)

    zs = np.concatenate([z * np.ones(nc) for z, nc in zip(zs, n_cells)])
    thetas = np.concatenate([theta * np.ones(nc) for theta, nc in zip(thetas, n_cells)])

    xs = a * np.cos(thetas) * np.cos(phis)
    ys = b * np.cos(thetas) * np.sin(phis)

    if pos_err > 0.0:
        xs += np.random.normal(scale=pos_err, size=thetas.shape)
        ys += np.random.normal(scale=pos_err, size=thetas.shape)
        zs += np.random.normal(scale=pos_err, size=thetas.shape)

    centers = np.vstack([xs, ys, zs]).T
    return centers


def ellipsoid_sheet(a, b, c, n_zs, **kwargs):
    """Creates an ellipsoidal apical mesh.

    Parameters
    ----------
    a, b, c : floats
       Size of the ellipsoid half axes in
       the x, y, and z directions, respectively
    n_zs : int
       The (approximate) number of faces along the z axis.

    kwargs are passed to `get_ellipsoid_centers`

    Returns
    -------
    eptm : a :class:`Epithelium` object

    The mesh returned is an `Epithelium` and not a simpler `Sheet`
    so that a unique cell data can hold information on the
    whole volume of the ellipsoid.

    """
    centers = get_ellipsoid_centers(a, b, c, n_zs, **kwargs)

    eptm = sheet_from_cell_centers(centers)
    eptm.settings["abc"] = [a, b, c]
    EllipsoidGeometry.update_all(eptm)
    return eptm


def spherical_sheet(radius, Nf, **kwargs):
    """Returns a spherical sheet with the given radius and (approximately)
    the given number of cells
    """

    n_zs = int(np.ceil(np.roots([2, 1.0, -Nf])[-1]))  # determined experimentaly ;p
    eptm = ellipsoid_sheet(radius, radius, radius, n_zs, **kwargs)
    eptm.settings.pop("abc")
    eptm.settings["radius"] = radius
    return eptm


def sheet_from_cell_centers(points, noise=0):
    """Returns a Sheet object from the Voronoï tessalation
    of the cell centers.

    The strategy is to project the points on a sphere, get the Voronoï
    tessalation on this sphere and reproject the vertices on the
    original (implicit) surface through linear interpolation of the cell centers.

    Works for relatively smooth surfaces (at the very minimum star convex).

    Parameters
    ----------

    points : np.ndarray of shape (Nf, 3)
        the x, y, z coordinates of the cell centers
    noise : float, default 0.0
        addiditve normal noise stdev

    Returns
    -------
    sheet : a :class:`Sheet` object with Nf faces


    """
    points = points.copy()
    if noise:
        points += np.random.normal(0, scale=noise, size=points.shape)
    points -= points.mean(axis=0)
    bbox = np.ptp(points, axis=0)
    points /= bbox

    rhos = np.linalg.norm(points, axis=1)
    thetas = np.arcsin(points[:, 2] / rhos)
    phis = np.arctan2(points[:, 0], points[:, 1])

    sphere_rad = rhos.max() * 1.1

    points_sphere = np.vstack(
        (
            sphere_rad * np.cos(thetas) * np.cos(phis),
            sphere_rad * np.cos(thetas) * np.sin(phis),
            sphere_rad * np.sin(thetas),
        )
    ).T
    points_sphere = np.concatenate(([[0, 0, 0]], points_sphere))

    vor3D = Voronoi(points_sphere)

    dsets = from_3d_voronoi(vor3D)
    eptm_ = Epithelium("v", dsets)

    eptm_ = single_cell(eptm_, 0)

    eptm = get_outer_sheet(eptm_)
    eptm.reset_index()
    eptm.reset_topo()
    eptm.vert_df["rho"] = np.linalg.norm(eptm.vert_df[eptm.coords], axis=1)
    mean_rho = eptm.vert_df["rho"].mean()

    SheetGeometry.scale(eptm, sphere_rad / mean_rho, ["x", "y", "z"])
    SheetGeometry.update_all(eptm)

    eptm.face_df["phi"] = np.arctan2(eptm.face_df.y, eptm.face_df.x)
    eptm.face_df["rho"] = np.linalg.norm(eptm.face_df[["x", "y", "z"]], axis=1)
    eptm.face_df["theta"] = np.arcsin(eptm.face_df.z / eptm.face_df["rho"])
    _itrp = interpolate.SmoothSphereBivariateSpline(
        thetas + np.pi / 2, phis + np.pi, rhos, s=1e-4
    )
    eptm.face_df["rho"] = _itrp(
        eptm.face_df["theta"] + np.pi / 2, eptm.face_df["phi"] + np.pi, grid=False
    )
    eptm.face_df["x"] = eptm.face_df.eval("rho * cos(theta) * cos(phi)")
    eptm.face_df["y"] = eptm.face_df.eval("rho * cos(theta) * sin(phi)")
    eptm.face_df["z"] = eptm.face_df.eval("rho * sin(theta)")

    eptm.edge_df[["fx", "fy", "fz"]] = eptm.upcast_face(eptm.face_df[["x", "y", "z"]])
    eptm.vert_df[["x", "y", "z"]] = eptm.edge_df.groupby("srce")[
        ["fx", "fy", "fz"]
    ].mean()
    for i, c in enumerate("xyz"):
        eptm.vert_df[c] *= bbox[i]

    SheetGeometry.update_all(eptm)

    eptm.sanitize(trim_borders=True)

    eptm.reset_index()
    eptm.reset_topo()
    SheetGeometry.update_all(eptm)
    null_length = eptm.edge_df.query("length == 0")

    while null_length.shape[0]:
        type1_transition(eptm, null_length.index[0])
        SheetGeometry.update_all(eptm)
        null_length = eptm.edge_df.query("length == 0")

    return eptm



"""
Dynamical models for monlayer and bulk epithelium.


"""

from . import effectors
from .factory import model_factory


BulkModel = model_factory(
    [
        effectors.LineTension,
        effectors.FaceContractility,
        effectors.CellAreaElasticity,
        effectors.CellVolumeElasticity,
    ],
    effectors.CellVolumeElasticity,
)


BulkModelwithFreeBorders = model_factory(
    [
        effectors.LineTension,
        effectors.FaceContractility,
        effectors.CellAreaElasticity,
        effectors.BorderElasticity,
        effectors.CellVolumeElasticity,
    ],
    effectors.CellVolumeElasticity,
)


class LaminaModel(BulkModel):
    """Not implemented yet
    """

    pass



from ..json_parser import load_spec
import os

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def quasistatic_bulk_spec():
    specfile = os.path.join(CURRENT_DIR, "bulk_qs.json")
    return load_spec(specfile)


def quasistatic_plane_spec():
    """Default specification for the dynamics
    of a sheet vertex model - also suitable for general faceted tissue

    .. code-block::

        {
        "face": {
            "contractility": 0.04,
            "vol_elasticity": 1.0,
            "prefered_height": 10.0,
            "prefered_area": 24.0,
            "prefered_vol": 0.0
            },
        "vert": {
             "radial_tension": 0.0
             },
        "edge": {
             "is_active": 1,
             "line_tension": 0.12
             },
        "settings": {
            "grad_norm_factor": 1.0,
            "nrj_norm_factor": 1.0
            }
        }

    """
    specfile = os.path.join(CURRENT_DIR, "plane_qs.json")
    return load_spec(specfile)


def quasistatic_sheet_spec():
    """Default specification for the dynamics
    of a sheet vertex model - also suitable for general faceted tissue

    .. code-block::

        {
        "face": {
            "contractility": 0.04,
            "vol_elasticity": 1.0,
            "prefered_height": 10.0,
            "prefered_area": 24.0,
            "prefered_vol": 0.0
            },
        "vert": {
             "radial_tension": 0.0
             },
        "edge": {
             "line_tension": 0.12
             },
        "settings": {
            "grad_norm_factor": 1.0,
            "nrj_norm_factor": 1.0
            }
        }

    """
    specfile = os.path.join(CURRENT_DIR, "sheet_qs.json")
    return load_spec(specfile)


def quasistatic_sheet_spec_anchors():
    """Default specification for the dynamics
    of a sheet vertex model with anchor vertices

    .. code-block::

        {
        "face": {
            "contractility": 0.04,
            "vol_elasticity": 1.0,
            "prefered_height": 10.0,
            "prefered_area": 24.0,
            "prefered_vol": 0.0
            },
        "vert": {
            "radial_tension": 0.0,
            "at_border": 0,
            "is_anchor": 0
             },
        "edge": {
            "anchor_elasticity": 0.0,
            "line_tension": 0.12,
            "at_border": 0,
            "is_anchor": 0
             },
        "settings": {
            "grad_norm_factor": 1.0,
            "nrj_norm_factor": 1.0
            }
        }

    """
    specfile = os.path.join(CURRENT_DIR, "sheet_qs_anchor.json")
    return load_spec(specfile)



"""
Event management module
=======================


"""

import logging
import random
from collections import deque
from datetime import datetime

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class EventManager:
    """
    Behavior management class based on two deques, the current and next one.



    """

    def __init__(self, element, logfile=None):
        """Creates an events class

        Parameters
        ----------
        element : str
           element on which the events occur, e.g face or cell
        logfile : str, default None
           if logfile is not None, will create a logging handler
           for this file where each event will be logged



        """
        self.current = deque()
        self.next = deque()
        self.element = element
        self.current.append((wait, {"face_id": -1, "n_steps": 1}))
        self.clock = 0
        if logfile is not None:
            fh = logging.FileHandler(logfile)
            fh.setLevel(logging.INFO)
            logger.addHandler(fh)
            logger.info("# Started logging at %s", datetime.now().isoformat())
            logger.info(f"time, {self.element} index, event")

    def extend(self, events):
        """
        Add a list of events to the next deque

        Parameters
        ----------
        events : list of tuples (behavior, [elem_id, args, kwargs])
            the three last elements don't need to be passed

        """
        for event in events:
            self.append(event[0], **event[1])

    def append(self, behavior, **kwargs):
        """Add an event to the manager's next deque

        behavior is a function whose signature is
        ..code :
            behavior(sheet, manager, **kwargs)

        this function itself might populate the managers next deque

        Parameters
        ----------
        behavior : function
        kwargs : dict defaults to {}
            keywords arguments to the behavior function
            if `"face_id"` is in the kwargs dictionnary,
            the face with this id is used.
        """
        if "face_id" in kwargs:
            elem_id = kwargs["face_id"]
        elif "elem_id" in kwargs:
            elem_id = kwargs["elem_id"]
        else:
            elem_id = -1

        for tup in self.next:
            unique = kwargs.get("unique", True)
            if "face_id" in tup[1]:
                if (elem_id == tup[1]["face_id"]) and (
                    behavior.__name__ == tup[0].__name__ and (unique)
                ):
                    return
            elif "elem_id" in tup[1]:
                if (elem_id == tup[1]["elem_id"]) and (
                    behavior.__name__ == tup[0].__name__(unique)
                ):
                    return
        self.next.append((behavior, kwargs))

    def execute(self, eptm):
        """
        Executes the events present in the `self.current` deque.
        """

        while self.current:
            (behavior, kwargs) = self.current.popleft()
            if "face_id" in kwargs:
                elem_id = kwargs["face_id"]
            elif "elem_id" in kwargs:
                elem_id = kwargs["elem_id"]
            else:
                elem_id = -1
            logger.info(f"{self.clock}, {elem_id}, {behavior.__name__}")
            behavior(eptm, self, **kwargs)

    def update(self):
        """
        Replaces `self.current` by `self.next` and clears `self.next`.
        """
        random.shuffle(self.next)
        self.current = self.next.copy()
        self.next.clear()


# Default dictionary for wait function
default_wait_spec = {"n_steps": 1}


def wait(eptm, manager, **kwargs):
    """Does nothing for a number of steps n_steps
    """
    wait_spec = default_wait_spec
    wait_spec.update(**kwargs)
    if kwargs["n_steps"] > 1:
        kwargs.update({"n_steps": kwargs["n_steps"] - 1})
        manager.next.append(("wait", wait_spec))



import pandas as pd

from ..utils.utils import _to_2d


def area_grad(sheet):

    coords = sheet.coords
    inv_area = sheet.edge_df.eval("1 / (4 * sub_area)")

    face_pos = sheet.edge_df[["f" + c for c in coords]]
    srce_pos = sheet.edge_df[["s" + c for c in coords]]
    trgt_pos = sheet.edge_df[["t" + c for c in coords]]

    r_ak = srce_pos - face_pos.values
    r_aj = trgt_pos - face_pos.values
    grad_a_srce = pd.DataFrame(index=sheet.edge_df.index, columns=["gx", "gy"])
    grad_a_srce["gx"] = r_aj["ty"] * sheet.edge_df["nz"]
    grad_a_srce["gy"] = -r_aj["tx"] * sheet.edge_df["nz"]
    grad_a_trgt = pd.DataFrame(index=sheet.edge_df.index, columns=["gx", "gy"])
    grad_a_trgt["gx"] = -r_ak["sy"] * sheet.edge_df["nz"]
    grad_a_trgt["gy"] = r_ak["sx"] * sheet.edge_df["nz"]

    grad_a_srce = _to_2d(inv_area) * grad_a_srce
    grad_a_trgt = _to_2d(inv_area) * grad_a_trgt

    return grad_a_srce, grad_a_trgt


def lumen_area_grad(eptm):
    apical_pos = eptm.vert_df[["x", "y"]].copy()
    apical_pos.loc[eptm.basal_verts] = 0
    srce_pos = eptm.upcast_srce(apical_pos)
    trgt_pos = eptm.upcast_trgt(apical_pos)
    grad_srce = srce_pos.copy()
    grad_srce.columns = ["gx", "gy"]
    grad_trgt = grad_srce.copy()
    grad_srce["gx"] = trgt_pos["y"]
    grad_srce["gy"] = -trgt_pos["x"]
    grad_trgt["gx"] = -srce_pos["y"]
    grad_trgt["gy"] = srce_pos["x"]
    # minus sign due to the backward orientation
    return -grad_srce, -grad_trgt



from ..json_parser import load_spec
import os

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def minimize_spec():
    """Solver default specification for scipy.optimize.minimize gradient
    descent

    .. code-block::

        {
            "norm_factor": 1,
            "minimize": {
                "method": "L-BFGS-B",
                "options": {"disp": false,
                            "ftol": 1e-6,
                            "gtol": 1e-3}
            }
        }

    """
    spec = {
        "norm_factor": 1,
        "minimize": {
            "method": "L-BFGS-B",
            "options": {"disp": False, "ftol": 1e-6, "gtol": 1e-3},
        },
    }
    return spec


def quasistatic():
    """Solver default specification for scipy.optimize.minimize gradient
    descent

    """
    spec = {
        "method": "L-BFGS-B",
        "options": {"disp": False, "ftol": 1e-6, "gtol": 1e-3},
    }
    return spec



print("import: 'tyssue'")
import tyssue

print("import: 'tyssue.behaviors'")
import tyssue.behaviors

print("import: 'tyssue.config'")
import tyssue.config

print("import: 'tyssue.core'")
import tyssue.core

print("import: 'tyssue.collisions'")
import tyssue.collisions

print("import: 'tyssue.draw'")
import tyssue.draw

print("import: 'tyssue.dynamics'")
import tyssue.dynamics

print("import: 'tyssue.geometry'")
import tyssue.geometry

print("import: 'tyssue.generation'")
import tyssue.generation

print("import: 'tyssue.io'")
import tyssue.io

print("import: 'tyssue.particles'")
import tyssue.particles

print("import: 'tyssue.solvers'")
import tyssue.solvers

print("import: 'tyssue.stores'")
import tyssue.stores

print("import: 'tyssue.topology'")
import tyssue.topology

print("import: 'tyssue.utils'")
import tyssue.utils




"""This module provides utlities to modify an input tissue through extrusion or subdivision
"""
import pandas as pd
import numpy as np


def extrude(apical_datasets, method="homotecy", scale=0.3, vector=[0, 0, -1]):
    """Extrude a sheet to form a monlayer epithelium

    Parameters
    ----------
    * apical_datasets: dictionnary of three DataFrames,
    'vert', 'edge', 'face'
    * method: str, optional {'homotecy'|'translation'|'normals'}
    default 'homotecy'
    * scale: float, optional
    the scale factor for homotetic scaling, default 0.3.
    * vector: sequence of three floats, optional,
    used for the translation

    default [0, 0, -1]

    if `method == 'homotecy'`, the basal layer is scaled down from the
    apical one homoteticaly w/r to the center of the coordinate
    system, by a factor given by `scale`

    if `method == 'translation'`, the basal vertices are translated from
    the apical ones by the vector `vect`

    if `method == 'normals'`, basal vertices are translated from
    the apical ones along the normal of the surface at each vertex,
    by a vector whose size is given by `scale`

    """
    apical_vert = apical_datasets["vert"]
    apical_face = apical_datasets["face"]
    apical_edge = apical_datasets["edge"]

    apical_vert["segment"] = "apical"
    apical_face["segment"] = "apical"
    apical_edge["segment"] = "apical"

    coords = list("xyz")
    datasets = {}

    Nv = apical_vert.index.max() + 1
    Ne = apical_edge.index.max() + 1
    Nf = apical_face.index.max() + 1

    basal_vert = apical_vert.copy()

    basal_vert.index = basal_vert.index + Nv
    basal_vert["segment"] = "basal"

    cell_df = apical_face[coords].copy()
    cell_df.index.name = "cell"
    cell_df["is_alive"] = 1

    basal_face = apical_face.copy()
    basal_face.index = basal_face.index + Nf
    basal_face[coords] = basal_face[coords] * 1 / 3.0
    basal_face["segment"] = "basal"
    basal_face["is_alive"] = 1

    apical_edge["cell"] = apical_edge["face"]
    basal_edge = apical_edge.copy()
    # ## Flip edge so that normals are outward
    basal_edge[["srce", "trgt"]] = basal_edge[["trgt", "srce"]] + Nv
    basal_edge["face"] = basal_edge["face"] + Nf
    basal_edge.index = basal_edge.index + Ne
    basal_edge["segment"] = "basal"

    lateral_face = pd.DataFrame(
        index=apical_edge.index + 2 * Nf, columns=apical_face.columns
    )
    lateral_face["segment"] = "lateral"
    lateral_face["is_alive"] = 1

    lateral_edge = pd.DataFrame(
        index=np.arange(2 * Ne, 6 * Ne), columns=apical_edge.columns
    )

    lateral_edge["cell"] = np.repeat(apical_edge["cell"].values, 4)
    lateral_edge["face"] = np.repeat(lateral_face.index.values, 4)
    lateral_edge["segment"] = "lateral"

    lateral_edge.loc[np.arange(2 * Ne, 6 * Ne, 4), ["srce", "trgt"]] = apical_edge[
        ["trgt", "srce"]
    ].values

    lateral_edge.loc[np.arange(2 * Ne + 1, 6 * Ne, 4), "srce"] = apical_edge[
        "srce"
    ].values
    lateral_edge.loc[np.arange(2 * Ne + 1, 6 * Ne, 4), "trgt"] = basal_edge[
        "trgt"
    ].values

    lateral_edge.loc[np.arange(2 * Ne + 2, 6 * Ne, 4), ["srce", "trgt"]] = basal_edge[
        ["trgt", "srce"]
    ].values

    lateral_edge.loc[np.arange(2 * Ne + 3, 6 * Ne, 4), "srce"] = basal_edge[
        "srce"
    ].values
    lateral_edge.loc[np.arange(2 * Ne + 3, 6 * Ne, 4), "trgt"] = apical_edge[
        "trgt"
    ].values

    if method == "homotecy":
        basal_vert[coords] = basal_vert[coords] * scale
    elif method == "translation":
        for c, u in zip(coords, vector):
            basal_vert[c] = basal_vert[c] + u
    elif method == "normals":
        field = apical_edge.groupby("srce")[["nx", "ny", "nz"]].mean()
        field = -field.values * scale / np.linalg.norm(field, axis=1)[:, None]
        basal_vert[coords] = basal_vert[coords] + field
    else:
        raise ValueError(
            """
        `method` argument not understood, supported values are 'homotecy'
        or 'translation'
        """
        )

    datasets["cell"] = cell_df
    datasets["vert"] = pd.concat([apical_vert, basal_vert])
    datasets["vert"]["is_active"] = 1
    datasets["edge"] = pd.concat([apical_edge, basal_edge, lateral_edge])
    datasets["face"] = pd.concat([apical_face, basal_face, lateral_face])
    datasets["edge"]["is_active"] = 1
    for elem in ["vert", "edge", "face", "cell"]:
        datasets[elem].index.name = elem

    return datasets


def create_anchors(sheet):
    """Adds an edge linked to every vertices at the boundary
    and create anchor vertices
    """
    anchor_specs = {
        "face": {"at_border": 0},
        "vert": {"at_border": 0, "is_anchor": 0},
        "edge": {"at_border": 0, "is_anchor": 0},
    }

    sheet.update_specs(anchor_specs)
    # ## Edges with no opposites denote the boundary

    free_edge = sheet.edge_df[sheet.edge_df["opposite"] == -1]
    free_vert = sheet.vert_df.loc[free_edge["srce"]]
    free_face = sheet.face_df.loc[free_edge["face"]]

    sheet.edge_df.loc[free_edge.index, "at_border"] = 1
    sheet.vert_df.loc[free_vert.index, "at_border"] = 1
    sheet.face_df.loc[free_face.index, "at_border"] = 1

    # ## Make a copy of the boundary vertices
    anchor_vert_df = free_vert.reset_index(drop=True)
    anchor_vert_df[sheet.coords] = anchor_vert_df[sheet.coords] * 1.01
    anchor_vert_df.index = anchor_vert_df.index + sheet.Nv
    anchor_vert_df["is_anchor"] = 1
    anchor_vert_df["at_border"] = 0
    anchor_vert_df["is_active"] = 0

    sheet.vert_df = pd.concat([sheet.vert_df, anchor_vert_df])
    sheet.vert_df.index.name = "vert"
    anchor_edge_df = pd.DataFrame(
        index=np.arange(sheet.Ne, sheet.Ne + free_vert.shape[0]),
        columns=sheet.edge_df.columns,
    )

    anchor_edge_df["srce"] = free_vert.index
    anchor_edge_df["trgt"] = anchor_vert_df.index
    anchor_edge_df["line_tension"] = 0
    anchor_edge_df["is_anchor"] = 1
    anchor_edge_df["face"] = -1
    anchor_edge_df["at_border"] = 0
    sheet.edge_df = pd.concat([sheet.edge_df, anchor_edge_df], sort=True)
    sheet.edge_df.index.name = "edge"
    sheet.reset_topo()


def subdivide_faces(eptm, faces):
    """Adds a vertex at the center of each face, and returns a
    new dataset

    Parameters
    ----------
    eptm: a :class:`Epithelium` instance
    faces: list,
     indices of the faces to be subdivided

    Returns
    -------
    new_dset: dict
      a dataset with the new faces devided

    """

    face_df = eptm.face_df.loc[faces]

    remaining = eptm.face_df.index.delete(faces)
    untouched_faces = eptm.face_df.loc[remaining]
    edge_df = pd.concat([eptm.edge_df[eptm.edge_df["face"] == face] for face in faces])
    verts = set(edge_df["srce"])
    vert_df = eptm.vert_df.loc[verts]

    Nsf = face_df.shape[0]
    Nse = edge_df.shape[0]

    eptm.vert_df["subdiv"] = 0
    untouched_faces["subdiv"] = 0
    eptm.edge_df["subdiv"] = 0

    new_vs_idx = pd.Series(np.arange(eptm.Nv, eptm.Nv + Nsf), index=face_df.index)
    upcast_new_vs = new_vs_idx.loc[edge_df["face"]].values

    new_vs = pd.DataFrame(
        index=pd.Index(np.arange(eptm.Nv, eptm.Nv + Nsf), name="vert"),
        columns=vert_df.columns,
    )
    new_es = pd.DataFrame(
        index=pd.Index(np.arange(eptm.Ne, eptm.Ne + 2 * Nse), name="edge"),
        columns=edge_df.columns,
    )

    new_vs["subdiv"] = 1
    # new_fs['subdiv'] = 1
    new_es["subdiv"] = 1
    if "cell" in edge_df.columns:
        new_es["cell"] = np.concatenate([edge_df["cell"], edge_df["cell"]])
    new_vs[eptm.coords] = face_df[eptm.coords].values
    # eptm.edge_df.loc[edge_df.index, 'face'] = new_fs.index
    # new_es['face'] = np.concatenate([new_fs.index,
    #                                  new_fs.index])
    new_es["face"] = np.concatenate([edge_df["face"], edge_df["face"]])
    new_es["srce"] = np.concatenate([edge_df["trgt"].values, upcast_new_vs])
    new_es["trgt"] = np.concatenate([upcast_new_vs, edge_df["srce"].values])
    new_dset = {
        "edge": pd.concat([eptm.edge_df, new_es]),
        "face": eptm.face_df,  # pd.concat([untouched_faces, new_fs]),
        "vert": pd.concat([eptm.vert_df, new_vs]),
    }

    if "cell" in edge_df.columns:
        new_dset["cell"] = eptm.cell_df

    return new_dset



import logging
import pandas as pd
from itertools import combinations

logger = logging.getLogger(name=__name__)


def add_vert(eptm, edge):
    """Adds a vertex in the middle of the edge,

    which is split as is its opposite(s)

    Parameters
    ----------
    eptm : a :class:`Epithelium` instance
    edge : int
    the index of one of the half-edges to split

    Returns
    -------
    new_vert : int
    the index to the new vertex
    new_edges : int or list of ints
    index to the new edge(s). For a sheet, returns
    a single index, for a 3D epithelium, returns
    the list of all the new parallel edges
    new_opp_edges : int or list of ints
    index to the new opposite edge(s). For a sheet, returns
    a single index, for a 3D epithelium, returns
    the list of all the new parallel edges


    In the simple case whith two half-edge, returns
    indices to the new edges, with the following convention:

    s    e    t
      ------>
    * <------ *
    oe

    s    e       ne   t
      ------   ----->
    * <----- * ------ *
        oe   nv   noe

    where "e" is the passed edge as argument, "s" its source "t" its
    target and "oe" its opposite. The returned edges are the ones
    between the new vertex and the input edge's original target.
    """

    srce, trgt = eptm.edge_df.loc[edge, ["srce", "trgt"]]
    opposites = eptm.edge_df[
        (eptm.edge_df["srce"] == trgt) & (eptm.edge_df["trgt"] == srce)
    ]
    parallels = eptm.edge_df[
        (eptm.edge_df["srce"] == srce) & (eptm.edge_df["trgt"] == trgt)
    ]

    new_vert = eptm.vert_df.loc[[srce, trgt]].mean()
    eptm.vert_df = eptm.vert_df.append(new_vert, ignore_index=True)
    new_vert = eptm.vert_df.index[-1]
    new_edges = []

    for p, p_data in parallels.iterrows():
        eptm.edge_df.loc[p, "trgt"] = new_vert
        eptm.edge_df = eptm.edge_df.append(p_data, ignore_index=True)
        new_edge = eptm.edge_df.index[-1]
        eptm.edge_df.loc[new_edge, "srce"] = new_vert
        eptm.edge_df.loc[new_edge, "trgt"] = trgt
        new_edges.append(new_edge)

    new_opp_edges = []
    for o, o_data in opposites.iterrows():
        eptm.edge_df.loc[o, "srce"] = new_vert
        eptm.edge_df = eptm.edge_df.append(o_data, ignore_index=True)
        new_opp_edge = eptm.edge_df.index[-1]
        eptm.edge_df.loc[new_opp_edge, "trgt"] = new_vert
        eptm.edge_df.loc[new_opp_edge, "srce"] = trgt
        new_opp_edges.append(new_opp_edge)

    # ## Sheet special case
    if len(new_edges) == 1:
        new_edges = new_edges[0]
    if len(new_opp_edges) == 1:
        new_opp_edges = new_opp_edges[0]
    elif len(new_opp_edges) == 0:
        new_opp_edges = None
    return new_vert, new_edges, new_opp_edges


def close_face(eptm, face):

    face_edges = eptm.edge_df[eptm.edge_df["face"] == face]
    srces = set(face_edges["srce"])
    trgts = set(face_edges["trgt"])

    if srces == trgts:
        logger.debug("Face {} already closed".format(face))
        return
    try:
        single_srce, = srces.difference(trgts)
        single_trgt, = trgts.difference(srces)
    except ValueError as err:
        print("Closing only possible with exactly two dangling vertices")
        raise err

    eptm.edge_df = eptm.edge_df.append(face_edges.iloc[0], ignore_index=True)
    eptm.edge_df.index.name = "edge"
    new_edge = eptm.edge_df.index[-1]
    eptm.edge_df.loc[new_edge, ["srce", "trgt"]] = single_trgt, single_srce


def condition_4i(eptm):
    """
    Return an index over the faces violating condition 4 i in Okuda et al 2013,
    that is edges (from the same face) sharing two vertices simultaneously.
    """
    num_srces = eptm.edge_df.groupby("face")["srce"].apply(lambda s: len(set(s)))
    num_sides = eptm.face_df["num_sides"]
    return eptm.face_df[num_srces != num_sides].index


def get_neighbour_face_pairs(eptm):
    """
    Returns a pandas Series of neighboring face pairs (as forzen sets of 2 indexes)
    """
    pairs = []
    eptm.edge_df["v_pair"] = eptm.edge_df[["srce", "trgt"]].apply(frozenset, axis=1)

    _ = eptm.edge_df.groupby("v_pair")["face"].apply(
        lambda s: pairs.extend(
            [frozenset((a, b)) for a, b in combinations(s.values, 2)]
        )
    )
    return pd.Series(pairs).drop_duplicates()


def get_num_common_edges(eptm):
    """
    Returns the number of common edges between two neighboring faces
    this number is set to -1 if those faces are opposite and share the
    same edges.
    """
    pairs = get_neighbour_face_pairs(eptm)
    face_v_pair_orbit = eptm.edge_df.groupby("face").apply(
        lambda df: frozenset(df["v_pair"])
    )
    n_common = [
        len(face_v_pair_orbit.loc[fa].intersection(face_v_pair_orbit.loc[fb]))
        if face_v_pair_orbit.loc[fb] != face_v_pair_orbit.loc[fa]
        else -1
        for fa, fb in pairs
    ]
    n_common = pd.Series(n_common, index=pd.Index(pairs, name="face_pairs"))
    return n_common


def condition_4ii(eptm):
    """
    Return a list of face pairs sharing more than one edge, as defined
    in Okuda et al. 2013 condition 4 ii
    """
    n_common = get_num_common_edges(eptm)
    return list(n_common[n_common >= 2].index)


def merge_border_edges(sheet):
    """Merge edges at the border of a sheet such that no vertex has only
    an on-going and an out-going edge.

    """

    single_trgt = sheet.edge_df[
        sheet.upcast_trgt(sheet.edge_df.groupby("trgt").apply(len) == 1)
    ]
    faces = set(single_trgt["face"])
    single_srce = sheet.edge_df[
        sheet.upcast_srce(sheet.edge_df.groupby("srce").apply(len) == 1)
    ]
    sheet.edge_df.drop(single_srce.index, inplace=True)
    sheet.edge_df.drop(
        set(single_trgt.index).difference(single_srce.index), inplace=True
    )
    for face in faces:
        close_face(sheet, face)
    sheet.reset_index()
    sheet.reset_topo()



import os
import warnings
from ..io import hdf5

stores_dir = os.path.abspath(os.path.dirname(__file__))
stores_list = os.listdir(stores_dir)

__doc__ = """Available predefined datasets:""" + "\n".join(stores_list)


def load_datasets(store, **kwargs):
    """
    This will soon be deprecated, use the leaner
    `tyssue.stores.stores_dir and tyssue.stores.stores_list
    """

    cur_dir = os.path.abspath(os.path.dirname(__file__))
    # root = os.path.dirname(cur_dir)
    warnings.warn(
        """
    This will soon be deprecated, use the leaner
    `tyssue.stores.stores_dir and tyssue.stores.stores_list
    """
    )
    filename = os.path.join(cur_dir, store)
    return hdf5.load_datasets(filename, **kwargs)



"""
Basic event module
=======================


"""

import logging
import numpy as np
from ...topology.sheet_topology import remove_face, type1_transition
from ...geometry.sheet_geometry import SheetGeometry

logger = logging.getLogger(__name__)


def grow(sheet, face, growth_rate):
    """Multiplies the equilibrium volume of face by a
    a factor (1+growth_rate)
    """
    sheet.face_df.loc[face, "prefered_vol"] *= 1 + growth_rate


def shrink(sheet, face, shrink_rate):
    """Devides the equilibrium volume of face face by a
    a factor 1+shrink_rate
    """
    sheet.face_df.loc[face, "prefered_vol"] /= 1 + shrink_rate


def exchange(sheet, face, geom, remove_tri_faces=True):
    """
    Execute a type1 transition on the shorter edge of a face.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of the face
    geom : a Geometry class
    """
    edges = sheet.edge_df[sheet.edge_df["face"] == face]
    shorter = edges.length.idxmin()
    # type1_transition(sheet, shorter, 2 * min(edges.length), remove_tri_faces)
    type1_transition(sheet, shorter, 0.1, remove_tri_faces)
    geom.update_all(sheet)


def remove(sheet, face, geom):
    """
    Removes the face and updates the geometry

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    face : index of the face
    geom : a Geometry class

    """
    remove_face(sheet, face)
    geom.update_all(sheet)


def contract(
    sheet,
    face,
    contractile_increase,
    multiple=False,
    contraction_column="contractility",
):
    """
    Contract the face by increasing the 'contractility' parameter
    by contractile_increase

    Parameters
    ----------
    face : id face

    """
    if multiple:
        sheet.face_df.loc[face, contraction_column] *= contractile_increase
    else:
        new_contractility = contractile_increase
        sheet.face_df.loc[face, contraction_column] += new_contractility


def ab_pull(sheet, face, radial_tension, distributed=False):
    """ Adds radial_tension to the face's vertices radial_tension
    """
    verts = sheet.edge_df[sheet.edge_df["face"] == face]["srce"].unique()
    if distributed:
        radial_tension = radial_tension / len(verts)

    sheet.vert_df.loc[verts, "radial_tension"] += radial_tension


def relax(sheet, face, contractility_decrease, contraction_column="contractility"):

    initial_contractility = 1.12
    new_contractility = (
        sheet.face_df.loc[face, contraction_column] / contractility_decrease
    )

    if new_contractility >= (initial_contractility / 2):
        sheet.face_df.loc[face, contraction_column] = new_contractility
        sheet.face_df.loc[face, "prefered_area"] *= contractility_decrease


def increase_linear_tension(sheet, face, line_tension, geom=SheetGeometry):
    edges = sheet.edge_df[sheet.edge_df["face"] == face]
    for index, edge in edges.iterrows():
        angle_ = np.arctan2(
            sheet.edge_df.loc[edge.name, "dx"], sheet.edge_df.loc[edge.name, "dy"]
        )

        if np.abs(angle_) < np.pi / 4:
            sheet.edge_df.loc[edge.name, "line_tension"] *= line_tension



import numpy as np
import pandas as pd


def make_df(index, spec):
    """

    """
    dtypes = np.dtype([(name, type(val)) for name, val in spec.items()])
    N = len(index)
    arr = np.empty(N, dtype=dtypes)
    df = pd.DataFrame.from_records(arr, index=index)
    for name, val in spec.items():
        df[name] = val
    return df



import numpy as np
from functools import wraps

from ..core.monolayer import Monolayer
from ..core.sheet import Sheet

from .base_topology import *
from .sheet_topology import type1_transition, remove_face
from .bulk_topology import HI_transition, IH_transition, find_HIs, find_IHs


MAX_ITER = 2


def auto_t1(fun):
    @wraps(fun)
    def with_rearange(*args, **kwargs):
        eptm, geom = args[:2]
        logger.debug("checking for t1s")
        res = fun(*args, **kwargs)
        shorts = find_IHs(eptm)
        if not len(shorts):
            return res
        i = 0
        while len(shorts):
            # shorter_edge = shorts[0]
            shorter_edge = np.random.choice(shorts)
            logger.info("transition on  edge %i", shorter_edge)
            if isinstance(eptm, Sheet):
                type1_transition(eptm, shorter_edge)
            else:
                IH_transition(eptm, shorter_edge)
            eptm.reset_index()
            eptm.reset_topo()
            geom.update_all(eptm)
            shorts = find_IHs(eptm)
            if len(shorts) and shorts[0] == shorter_edge:
                # IH transition did not work, skipping
                logger.info("removing shorter edge %i from edge list", shorter_edge)
                shorts = shorts[1:]
            i += 1
            if i > MAX_ITER:
                break
        if eptm.position_buffer is not None:
            logger.info("out T1 changed buffer")
            eptm.position_buffer = eptm.vert_df[eptm.coords].copy()
        logger.info("performed %i T1", i)
        return res

    return with_rearange


def auto_t3(fun):
    @wraps(fun)
    def with_rearange(*args, **kwargs):
        eptm, geom = args[:2]
        logger.debug("checking for t3s")
        res = fun(*args, **kwargs)
        tri_faces = find_HIs(eptm)
        if not len(tri_faces):
            return res
        i = 0
        while len(tri_faces):
            smaller_face = tri_faces[0]
            logger.debug("Performing t3 on face %d", smaller_face)
            if isinstance(eptm, Sheet):
                remove_face(eptm, smaller_face)
            else:
                HI_transition(eptm, smaller_face)
            eptm.reset_index()
            eptm.reset_topo()
            geom.update_all(eptm)
            tri_faces = find_HIs(eptm)
            i += 1
            if i > MAX_ITER:
                break
        if eptm.position_buffer is not None:
            logger.info("out T3 changed buffer")
            eptm.position_buffer = eptm.vert_df[eptm.coords].copy()
        logger.info("performed %i T3", i)
        return res

    return with_rearange



import numpy as np
import pandas as pd
import warnings
import logging

log = logging.getLogger(name=__name__)

from scipy import optimize

from ..utils import scaled_unscaled


def bruteforce_isotropic_relax(eptm, geom, model):
    def to_minimize(scale):
        return scaled_unscaled(model.compute_energy, scale, eptm, geom, args=[eptm])

    optim_res = optimize.minimize_scalar(
        to_minimize, method="bounded", bounds=[1e-6, 1e2]
    )
    if optim_res["success"]:
        log.info("Scaling by factor {:.3f}".format(optim_res["x"]))
        scale = optim_res["x"]
        geom.scale(eptm, scale, eptm.coords)
        geom.update_all(eptm)
    else:
        warnings.warn("Optimisation failed")

    return optim_res



import logging
import warnings
import itertools
from functools import wraps


import numpy as np
import pandas as pd

from .sheet_topology import face_division
from .base_topology import add_vert, close_face
from ..geometry.utils import rotation_matrix
from ..geometry.bulk_geometry import BulkGeometry
from ..core.objects import get_opposite_faces
from ..core.monolayer import Monolayer

logger = logging.getLogger(name=__name__)
MAX_ITER = 10


def get_division_edges(eptm, mother, plane_normal, plane_center=None):
    """
    """
    plane_normal = np.asarray(plane_normal)
    if plane_center is None:
        plane_center = eptm.cell_df.loc[mother, eptm.coords]

    n_xy = np.linalg.norm(plane_normal[:2])
    theta = -np.arctan2(n_xy, plane_normal[2])
    if np.linalg.norm(plane_normal[:2]) < 1e-10:
        rot = None
    else:
        direction = [plane_normal[1], -plane_normal[0], 0]
        rot = rotation_matrix(theta, direction)
    cell_verts = set(eptm.edge_df[eptm.edge_df["cell"] == mother]["srce"])
    vert_pos = eptm.vert_df.loc[cell_verts, eptm.coords]
    for c in eptm.coords:
        vert_pos[c] -= plane_center[c]
    if rot is not None:
        vert_pos[:] = np.dot(vert_pos, rot)

    mother_edges = eptm.edge_df[eptm.edge_df["cell"] == mother]
    srce_z = vert_pos.loc[mother_edges["srce"], "z"]
    srce_z.index = mother_edges.index
    trgt_z = vert_pos.loc[mother_edges["trgt"], "z"]
    trgt_z.index = mother_edges.index
    division_edges = mother_edges[((srce_z < 0) & (trgt_z >= 0))]

    # Order the returned edges so that their centers
    # are oriented counterclockwize in the division plane
    # in preparation for septum creation
    srce_pos = vert_pos.loc[division_edges["srce"], eptm.coords].values
    trgt_pos = vert_pos.loc[division_edges["trgt"], eptm.coords].values
    centers = (srce_pos + trgt_pos) / 2
    theta = np.arctan2(centers[:, 1], centers[:, 0])
    return division_edges.iloc[np.argsort(theta)].index


def get_division_vertices(
    eptm, division_edges=None, mother=None, plane_normal=None, plane_center=None
):

    if division_edges is None:
        division_edges = get_division_edges(eptm, mother, plane_normal, plane_center)
    vertices = []
    for edge in division_edges:
        new_vert, *_ = add_vert(eptm, edge)
        vertices.append(new_vert)
    return vertices


def cell_division(eptm, mother, geom, vertices=None):

    if vertices is None:
        vertices = get_division_vertices(
            eptm,
            division_edges=None,
            mother=mother,
            plane_normal=None,
            plane_center=None,
        )

    cell_cols = eptm.cell_df.loc[mother]
    eptm.cell_df = eptm.cell_df.append(cell_cols, ignore_index=True)
    eptm.cell_df.index.name = "cell"
    daughter = eptm.cell_df.index[-1]

    pairs = set(
        [
            frozenset([v1, v2])
            for v1, v2 in itertools.product(vertices, vertices)
            if v1 != v2
        ]
    )
    daughter_faces = []

    # devide existing faces
    for v1, v2 in pairs:
        v1_faces = eptm.edge_df[eptm.edge_df["srce"] == v1]["face"]
        v2_faces = eptm.edge_df[eptm.edge_df["srce"] == v2]["face"]
        # we should devide a face if both v1 and v2
        # are part of it
        faces = set(v1_faces).intersection(v2_faces)
        for face in faces:
            daughter_faces.append(face_division(eptm, face, v1, v2))
    # septum
    face_cols = eptm.face_df.iloc[-2:]
    eptm.face_df = eptm.face_df.append(face_cols, ignore_index=True)
    eptm.face_df.index.name = "face"
    septum = eptm.face_df.index[-2:]
    daughter_faces.extend(list(septum))

    num_v = len(vertices)
    num_new_edges = num_v * 2

    edge_cols = eptm.edge_df.iloc[-num_new_edges:]
    eptm.edge_df = eptm.edge_df.append(edge_cols, ignore_index=True)
    eptm.edge_df.index.name = "edge"
    new_edges = eptm.edge_df.index[-num_new_edges:]

    # To keep mother orientation, the first septum face
    # belongs to mother
    for v1, v2, edge, oppo in zip(
        vertices, np.roll(vertices, -1), new_edges[:num_v], new_edges[num_v:]
    ):
        # Mother septum
        eptm.edge_df.loc[edge, ["srce", "trgt", "face", "cell"]] = (
            v1,
            v2,
            septum[0],
            mother,
        )
        # Daughter septum
        eptm.edge_df.loc[oppo, ["srce", "trgt", "face", "cell"]] = (
            v2,
            v1,
            septum[1],
            daughter,
        )

    eptm.reset_index()
    eptm.reset_topo()
    geom.update_all(eptm)

    m_septum_edges = eptm.edge_df[eptm.edge_df["face"] == septum[0]]
    m_septum_norm = m_septum_edges[eptm.ncoords].mean()
    m_septum_pos = eptm.face_df.loc[septum[0], eptm.coords]

    # splitting the faces between mother and daughter
    # based on the orientation of the vector from septum
    # center to each face center w/r to the septum norm
    mother_faces = set(eptm.edge_df[eptm.edge_df["cell"] == mother]["face"])
    for face in mother_faces:
        if face == septum[0]:
            continue

        dr = eptm.face_df.loc[face, eptm.coords] - m_septum_pos
        proj = (dr.values * m_septum_norm).sum(axis=0)
        f_edges = eptm.edge_df[eptm.edge_df["face"] == face].index
        if proj < 0:
            eptm.edge_df.loc[f_edges, "cell"] = mother
        else:
            eptm.edge_df.loc[f_edges, "cell"] = daughter
    eptm.reset_index()
    eptm.reset_topo()
    return daughter


def find_rearangements(eptm):
    """Finds the candidates for IH and HI transitions
    Returns
    -------
    edges_HI: set of indexes of short edges
    faces_IH: set of indexes of small triangular faces
    """
    l_th = eptm.settings.get("threshold_length", 1e-6)
    shorts = eptm.edge_df[eptm.edge_df["length"] < l_th]
    if not shorts.shape[0]:
        return []
    edges_IH = find_IHs(eptm, shorts)
    faces_HI = find_HIs(eptm, shorts)
    return edges_IH, faces_HI


def find_IHs(eptm, shorts=None):

    l_th = eptm.settings.get("threshold_length", 1e-6)
    if shorts is None:
        shorts = eptm.edge_df[eptm.edge_df["length"] < l_th]
    if not shorts.shape[0]:
        return []

    edges_IH = shorts.groupby(["srce", "trgt"]).apply(
        lambda df: pd.Series(
            {
                "edge": df.index[0],
                "length": df["length"].iloc[0],
                "num_sides": min(eptm.face_df.loc[df["face"], "num_sides"]),
                "pair": frozenset(df.iloc[0][["srce", "trgt"]]),
            }
        )
    )
    # keep only one of the edges per vertex pair and sort by length
    edges_IH = (
        edges_IH[edges_IH["num_sides"] > 3]
        .drop_duplicates("pair")
        .sort_values("length")
    )
    return edges_IH["edge"].values


def find_HIs(eptm, shorts=None):
    l_th = eptm.settings.get("threshold_length", 1e-6)
    if shorts is None:
        shorts = eptm.edge_df[(eptm.edge_df["length"] < l_th)]
    if not shorts.shape[0]:
        return []

    max_f_length = shorts.groupby("face")["length"].apply(max)
    short_faces = eptm.face_df.loc[max_f_length[max_f_length < l_th].index]
    faces_HI = short_faces[short_faces["num_sides"] == 3].sort_values("area").index
    return faces_HI


def IH_transition(eptm, e_1011):
    """
    I → H transition as defined in Okuda et al. 2013
    (DOI 10.1007/s10237-012-0430-7).
    See tyssue/doc/illus/IH_transition.png for the definition of the
    edges, which follow the one in the above article
    """

    v10, v11 = eptm.edge_df.loc[e_1011, ["srce", "trgt"]]
    v_pairs = _get_vertex_pairs_IH(eptm, e_1011)
    if v_pairs is None:
        logger.warning(
            "Edge %i is not a valid junction to perform IH transition, aborting", e_1011
        )
        return
    try:
        (v1, v4), (v2, v5), (v3, v6) = v_pairs
    except ValueError as err:
        logger.warning(
            "Edge %i is not a valid junction to perform IH transition, aborting", e_1011
        )
        return

    if len({v1, v4, v2, v5, v3, v6}) != 6:
        logger.warning(
            "Edge %i has adjacent triangular faces"
            " can't perform IH transition, aborting",
            e_1011,
        )
        return

    new_vs = eptm.vert_df.loc[[v1, v2, v3]].copy()
    eptm.vert_df = eptm.vert_df.append(new_vs, ignore_index=True)
    v7, v8, v9 = eptm.vert_df.index[-3:]

    cells = []
    srce_cell_orbits = eptm.get_orbits("srce", "cell")
    for vi, vj, vk in [
        (v1, v2, v3),
        (v4, v5, v6),
        (v1, v2, v11),
        (v2, v3, v11),
        (v3, v1, v11),
    ]:
        cell = list(
            set(srce_cell_orbits.loc[vi])
            .intersection(srce_cell_orbits.loc[vj])
            .intersection(srce_cell_orbits.loc[vk])
        )
        cells.append(cell[0] if cell else None)

    cA, cB, cC, cD, cE = cells
    if cA is not None:
        # orient vertices 1,2,3 positively
        r_12 = (
            eptm.vert_df.loc[v2, eptm.coords].values
            - eptm.vert_df.loc[v1, eptm.coords].values
        ).astype(np.float)
        r_23 = (
            eptm.vert_df.loc[v3, eptm.coords].values
            - eptm.vert_df.loc[v2, eptm.coords].values
        ).astype(np.float)
        r_123 = eptm.vert_df.loc[[v1, v2, v3], eptm.coords].mean(axis=0).values
        r_A = eptm.cell_df.loc[cA, eptm.coords].values
        orient = np.dot(np.cross(r_12, r_23), (r_123 - r_A))
    elif cB is not None:
        # orient vertices 4,5,6 negatively
        r_45 = (
            eptm.vert_df.loc[v5, eptm.coords].values
            - eptm.vert_df.loc[v4, eptm.coords].values
        ).astype(np.float)
        r_56 = (
            eptm.vert_df.loc[v6, eptm.coords].values
            - eptm.vert_df.loc[v5, eptm.coords].values
        ).astype(np.float)
        r_456 = eptm.vert_df.loc[[v4, v5, v6], eptm.coords].mean(axis=0).values
        r_B = eptm.cell_df.loc[cB, eptm.coords].values
        orient = -np.dot(np.cross(r_45, r_56), (r_456 - r_B))
    else:
        logger.warning(
            "I - H transition is not possible without cells on either ends"
            " of the edge - would result in a hole"
        )
        return

    if orient < 0:
        v1, v2, v3 = v1, v3, v2
        v4, v5, v6 = v4, v6, v5
        cC, cE = cE, cC
    vertices = [v1, v2, v3, v4, v5, v6, v7, v8, v9, v10, v11]

    for i, va, vb, new in zip(range(3), (v1, v2, v3), (v4, v5, v6), (v7, v8, v9)):
        # assign v1 -> v10 edges to  v1 -> v7
        e_a10s = eptm.edge_df[
            (eptm.edge_df["srce"] == va) & (eptm.edge_df["trgt"] == v10)
        ].index
        eptm.edge_df.loc[e_a10s, "trgt"] = new
        # assign v10 -> v1 edges to  v7 -> v1
        e_10as = eptm.edge_df[
            (eptm.edge_df["srce"] == v10) & (eptm.edge_df["trgt"] == va)
        ].index
        eptm.edge_df.loc[e_10as, "srce"] = new
        # assign v4 -> v11 edges to  v4 -> v7
        e_b11s = eptm.edge_df[
            (eptm.edge_df["srce"] == vb) & (eptm.edge_df["trgt"] == v11)
        ].index
        eptm.edge_df.loc[e_b11s, "trgt"] = new
        # assign v11 -> v4 edges to  v7 -> v4
        e_11bs = eptm.edge_df[
            (eptm.edge_df["srce"] == v11) & (eptm.edge_df["trgt"] == vb)
        ].index
        eptm.edge_df.loc[e_11bs, "srce"] = new

    _set_new_pos_IH(eptm, e_1011, vertices)

    face = eptm.edge_df.loc[e_1011, "face"]
    new_fs = eptm.face_df.loc[[face, face]].copy()
    eptm.face_df = eptm.face_df.append(new_fs, ignore_index=True)
    fa, fb = eptm.face_df.index[-2:]
    edges_fa_fb = eptm.edge_df.loc[[e_1011] * 6].copy()
    eptm.edge_df = eptm.edge_df.append(edges_fa_fb, ignore_index=True)
    new_es = eptm.edge_df.index[-6:]
    for eA, eB, (vi, vj) in zip(
        new_es[::2], new_es[1::2], [(v7, v8), (v8, v9), (v9, v7)]
    ):
        eptm.edge_df.loc[eA, ["srce", "trgt", "face", "cell"]] = vi, vj, fa, cA
        eptm.edge_df.loc[eB, ["srce", "trgt", "face", "cell"]] = vj, vi, fb, cB

    for cell in cells:
        for face in eptm.edge_df[eptm.edge_df["cell"] == cell]["face"]:
            close_face(eptm, face)

    # Removing the remaining edges and vertices
    todel_edges = eptm.edge_df[
        (eptm.edge_df["srce"] == v10)
        | (eptm.edge_df["trgt"] == v10)
        | (eptm.edge_df["srce"] == v11)
        | (eptm.edge_df["trgt"] == v11)
        | pd.isna(eptm.edge_df["cell"])
    ].index

    eptm.edge_df = eptm.edge_df.loc[eptm.edge_df.index.delete(todel_edges)]
    eptm.vert_df = eptm.vert_df.loc[set(eptm.edge_df.sort_values("srce")["srce"])]
    eptm.face_df = eptm.face_df.loc[set(eptm.edge_df.sort_values("face")["face"])]
    eptm.cell_df = eptm.cell_df.loc[set(eptm.edge_df.sort_values("cell")["cell"])]

    eptm.edge_df.index.name = "edge"
    if isinstance(eptm, Monolayer):
        for vert in (v7, v8, v9):
            eptm.guess_vert_segment(vert)
        for face in fa, fb:
            eptm.guess_face_segment(face)

    eptm.reset_index()
    eptm.reset_topo()


def HI_transition(eptm, face):
    """
    H → I transition as defined in Okuda et al. 2013
    (DOI 10.1007/s10237-012-0430-7).
    See tyssue/doc/illus/IH_transition.png for the definition of the
    edges, which follow the one in the above article
    """
    if eptm.face_df.loc[face, "num_sides"] != 3:
        raise ValueError("Only three sided faces can undergo a H-I transition")

    fa = face
    f_edges = eptm.edge_df[eptm.edge_df["face"] == face]
    v7 = f_edges.iloc[0]["srce"]
    v8 = f_edges.iloc[0]["trgt"]
    v9 = f_edges[f_edges["srce"] == v8]["trgt"].iloc[0]

    cA = f_edges["cell"].iloc[0]

    eptm.get_opposite_faces()
    fb = eptm.face_df["opposite"].loc[face]
    if fb > 0:
        cB = eptm.edge_df[eptm.edge_df["face"] == fb]["cell"].iloc[0]
    else:
        cB = None

    cA_edges = eptm.edge_df[eptm.edge_df["cell"] == cA]

    v_pairs = []
    for vk in (v7, v8, v9):
        vis = set(cA_edges[cA_edges["srce"] == vk]["trgt"])
        try:
            vi, = vis.difference({v7, v8, v9})
        except ValueError:
            warnings.warn("Invalid topology for a HI transition, aborting")
            return
        vjs = set(eptm.edge_df[eptm.edge_df["srce"] == vk]["trgt"])
        try:
            vj, = vjs.difference({v7, v8, v9, vi})
        except ValueError:
            warnings.warn("Invalid topology for a HI transition, aborting")
            return
        v_pairs.append((vi, vj))

    (v1, v4), (v2, v5), (v3, v6) = v_pairs

    srce_cell_orbit = eptm.get_orbits("srce", "cell")
    cells = [cA, cB]
    for (vi, vj, vk) in [(v1, v2, v4), (v2, v3, v5), (v1, v3, v4)]:
        cell = list(
            set(srce_cell_orbit.loc[vi])
            .intersection(srce_cell_orbit.loc[vj])
            .intersection(srce_cell_orbit.loc[vk])
        )

        cells.append(cell[0] if cell else None)

    cA, cB, cC, cD, cE = cells

    eptm.vert_df = eptm.vert_df.append(eptm.vert_df.loc[[v8, v9]], ignore_index=True)
    eptm.vert_df.index.name = "vert"
    v10, v11 = eptm.vert_df.index[-2:]
    _set_new_pos_HI(eptm, fa, v10, v11)

    for vi, vj, vk in zip((v1, v2, v3), (v4, v5, v6), (v7, v8, v9)):
        e_iks = eptm.edge_df[
            (eptm.edge_df["srce"] == vi) & (eptm.edge_df["trgt"] == vk)
        ].index
        eptm.edge_df.loc[e_iks, "trgt"] = v10

        e_kis = eptm.edge_df[
            (eptm.edge_df["srce"] == vk) & (eptm.edge_df["trgt"] == vi)
        ].index
        eptm.edge_df.loc[e_kis, "srce"] = v10

        e_jks = eptm.edge_df[
            (eptm.edge_df["srce"] == vj) & (eptm.edge_df["trgt"] == vk)
        ].index
        eptm.edge_df.loc[e_jks, "trgt"] = v11

        e_kjs = eptm.edge_df[
            (eptm.edge_df["srce"] == vk) & (eptm.edge_df["trgt"] == vj)
        ].index
        eptm.edge_df.loc[e_kjs, "srce"] = v11

    # Closing the faces with v10 → v11 edges
    for cell in cells:
        for face in eptm.edge_df[eptm.edge_df["cell"] == cell]["face"]:
            close_face(eptm, face)

    # Removing the remaining edges and vertices
    todel_edges = eptm.edge_df[
        (eptm.edge_df["srce"] == v7)
        | (eptm.edge_df["trgt"] == v7)
        | (eptm.edge_df["srce"] == v8)
        | (eptm.edge_df["trgt"] == v8)
        | (eptm.edge_df["srce"] == v9)
        | (eptm.edge_df["trgt"] == v9)
    ].index

    eptm.edge_df = eptm.edge_df.loc[eptm.edge_df.index.delete(todel_edges)]
    eptm.vert_df = eptm.vert_df.loc[eptm.vert_df.index.delete([v7, v8, v9])]
    orphan_faces = set(eptm.face_df.index).difference(eptm.edge_df.face)
    eptm.face_df = eptm.face_df.loc[
        eptm.face_df.index.delete(list(orphan_faces))
    ].copy()
    eptm.edge_df.index.name = "edge"
    eptm.reset_index()
    eptm.reset_topo()


def _add_edge_to_existing(eptm, cell, vi, vj, new_srce, new_trgt):
    """
    Add edges between vertices v7, v8 and v9 to the existing faces
    """
    cell_edges = eptm.edge_df[eptm.edge_df["cell"] == cell]
    for f, data in cell_edges.groupby("face"):
        if {vi, vj, new_srce, new_trgt}.issubset(set(data["srce"]).union(data["trgt"])):
            good_f = f
            break
    else:
        raise ValueError(
            "no face with vertices {}, {}, {} and {}"
            " was found for cell {}".format(vi, vj, new_srce, new_trgt, cell)
        )
    eptm.edge_df = eptm.edge_df.append(cell_edges.iloc[-1], ignore_index=True)
    new_e = eptm.edge_df.index[-1]
    eptm.edge_df.loc[new_e, ["srce", "trgt", "face", "cell"]] = (
        new_srce,
        new_trgt,
        good_f,
        cell,
    )


def _set_new_pos_IH(eptm, e_1011, vertices):
    """Okuda 2013 equations 46 to 56
    """
    Dl_th = eptm.settings["threshold_length"]

    (v1, v2, v3, v4, v5, v6, v7, v8, v9, v10, v11) = vertices

    # eq. 49
    r_1011 = -eptm.edge_df.loc[e_1011, eptm.dcoords].values
    u_T = r_1011 / np.linalg.norm(r_1011)
    # eq. 50
    r0 = eptm.vert_df.loc[[v10, v11], eptm.coords].mean(axis=0).values

    v_0ns = []
    for vi, vj, vk in zip((v1, v2, v3), (v4, v5, v6), (v7, v8, v9)):
        # eq. 54 - 56
        r0i, r0j = eptm.vert_df.loc[[vi, vj], eptm.coords].values - r0[np.newaxis, :]
        w_0k = (r0i / np.linalg.norm(r0i) + r0j / np.linalg.norm(r0j)) / 2
        # eq. 51 - 53
        v_0k = w_0k - (np.dot(w_0k, u_T)) * u_T
        v_0ns.append(v_0k)

    # see definition of l_max bellow eq. 56
    l_max = np.max(
        [np.linalg.norm(v_n - v_m) for (v_n, v_m) in itertools.combinations(v_0ns, 2)]
    )
    # eq. 46 - 49
    for vk, v_0k in zip((v7, v8, v9), v_0ns):
        eptm.vert_df.loc[vk, eptm.coords] = r0 + (Dl_th / l_max) * v_0k


def _get_vertex_pairs_IH(eptm, e_1011):

    srce_face_orbits = eptm.get_orbits("srce", "face")
    v10, v11 = eptm.edge_df.loc[e_1011, ["srce", "trgt"]]
    v10_out = set(eptm.edge_df[eptm.edge_df["srce"] == v10]["trgt"]) - {v11}
    faces_123 = {v: set(srce_face_orbits.loc[v]) for v in v10_out}

    v11_out = set(eptm.edge_df[eptm.edge_df["srce"] == v11]["trgt"]) - {v10}
    faces_456 = {v: set(srce_face_orbits.loc[v]) for v in v11_out}
    v_pairs = []
    for vi in v10_out:
        for vj in v11_out:
            common_face = faces_123[vi].intersection(faces_456[vj])
            if common_face:
                v_pairs.append((vi, vj))
                break
        else:
            return None
    return v_pairs


def _set_new_pos_HI(eptm, fa, v10, v11):

    r0 = eptm.face_df.loc[fa, eptm.coords].values

    norm_a = eptm.edge_df[eptm.edge_df["face"] == fa][eptm.ncoords].mean(axis=0).values
    norm_a = norm_a / np.linalg.norm(norm_a)
    norm_b = -norm_a
    Dl_th = eptm.settings["threshold_length"] * 1.01
    eptm.vert_df.loc[v10, eptm.coords] = r0 + Dl_th / 2 * norm_b
    eptm.vert_df.loc[v11, eptm.coords] = r0 + Dl_th / 2 * norm_a



import warnings
import pandas as pd
import numpy as np

from collections import defaultdict
from .sheet import Sheet
from .objects import Epithelium


def _filter_columns(cols_hist, cols_in, element):
    if not set(cols_hist).issubset(cols_in):
        warnings.warn(
            f"""
Columns {set(cols_hist).difference(cols_in)} are in the history
 {element} dataframe but not in the sheet {element} dataframe.

These non existent columns will not be saved."""
        )
        cols_hist = set(cols_hist).intersection(cols_in)
    return list(cols_hist)


class History:
    """ This class handles recording and retrieving time series
    of sheet objects.

    """

    def __init__(self, sheet, extra_cols=None):
        """Creates a `SheetHistory` instance.

        Parameters
        ----------
        sheet : a :class:`Sheet` object which we want to record
        extra_cols : dictionnary with sheet.datasets as keys and list of
            columns as values. Default None
        """
        if extra_cols is None:
            extra_cols = defaultdict(list)
        else:
            extra_cols = defaultdict(list, **extra_cols)

        self.sheet = sheet
        self.time_id = 0

        self.datasets = {}
        self.columns = {}
        vcols = sheet.coords + extra_cols["vert"]
        self.vcols = _filter_columns(vcols, sheet.vert_df.columns, "vertex")
        self.vert_h = sheet.vert_df[self.vcols].reset_index(drop=False)
        if not "time_id" in self.vcols:
            self.vert_h["time_id"] = 0
        self.datasets["vert"] = self.vert_h
        self.columns["vert"] = self.vcols

        fcols = extra_cols["face"]
        self.fcols = _filter_columns(fcols, sheet.face_df.columns, "face")
        self.face_h = sheet.face_df[self.fcols].reset_index(drop=False)
        if not "time_id" in self.fcols:
            self.face_h["time_id"] = 0
        self.datasets["face"] = self.face_h
        self.columns["face"] = self.fcols

        if sheet.cell_df is not None:
            ccols = extra_cols["cell"]
            self.ccols = _filter_columns(ccols, sheet.cell_df.columns, "cell")
            self.cell_h = sheet.cell_df[self.ccols].reset_index(drop=False)
            if not "time_id" in self.ccols:
                self.cell_h["time_id"] = 0
            self.datasets["cell"] = self.cell_h
            self.columns["cell"] = self.ccols
            extra_cols["edge"].append("cell")

        ecols = ["srce", "trgt", "face"] + extra_cols["edge"]
        self.ecols = _filter_columns(ecols, sheet.edge_df.columns, "edge")
        self.edge_h = sheet.edge_df[self.ecols].reset_index(drop=False)
        if not "time_id" in self.ecols:
            self.edge_h["time_id"] = 0
        self.datasets["edge"] = self.edge_h
        self.columns["edge"] = self.ecols
        self.time_index = []
        self.time_stamps = []

    def record(self, to_record=["vert"], time_stamp=None):
        """Appends a copy of the sheet datasets to the history instance.

        Parameters
        ----------
        to_report : list of strings, default ['vert']
            the datasets from self.sheet to be saved

        """
        self.time_id += 1
        self.time_index.append(self.time_id)

        if time_stamp is not None:
            self.time_stamps.append(time_stamp)

        for element in to_record:
            hist = self.datasets[element]
            cols = self.columns[element]
            df = self.sheet.datasets[element][cols].reset_index(drop=False)
            if not "time_id" in cols:
                times = pd.Series(
                    np.ones((df.shape[0],)) * self.time_id, name="time_id"
                )
                df = pd.concat([df, times], ignore_index=False, axis=1, sort=False)
            hist = pd.concat([hist, df], ignore_index=True, axis=0, sort=False)
            self.datasets[element] = hist

    def retrieve(self, index):
        """Return datasets at time `time`.

        If a specific dataset was not recorded at time time, the closest record before that
        time is used.
        """
        sheet_datasets = {}
        for element in self.datasets:
            hist = self.datasets[element]
            cols = self.columns[element]
            df = _retrieve(hist, index)
            df = df.set_index(element)[cols]
            sheet_datasets[element] = df

        return sheet_datasets

    def __iter__(self):

        for t in self.time_index:
            sheet = type(self.sheet)(
                f"{self.sheet.identifier}_{t:04d}", self.retrieve(t), self.sheet.specs
            )
            yield t, sheet


def _retrieve(dset, time_id):
    times = dset["time_id"].values
    t = times[times <= time_id][-1]
    return dset[dset["time_id"] == t]



import logging
import numpy as np
import pandas as pd
import warnings
from functools import wraps

from .intersection import self_intersections
from ..core.sheet import Sheet, get_outer_sheet

log = logging.getLogger(__name__)


def auto_collisions(fun):
    """Decorator to solve collisions detections after the
    execution of the decorated function.

    It is assumed that the two first arguments of the decorated
    function are a :class:`Sheet` object and a geometry class

    Note
    ----
    The function is re-executed with the updated geometry
    """

    @wraps(fun)
    def with_collision_correction(*args, **kwargs):
        log.debug("checking for collisions")
        eptm, geom = args[:2]
        eptm.position_buffer = eptm.vert_df[eptm.coords].copy()
        res = fun(*args, **kwargs)
        if isinstance(eptm, Sheet):
            change = solve_sheet_collisions(eptm, eptm.position_buffer)
        else:
            change = solve_bulk_collisions(eptm, eptm.position_buffer)
        if change:
            log.info("collision avoided")
        geom.update_all(eptm)
        return res

    return with_collision_correction


def solve_bulk_collisions(eptm, position_buffer):
    """Corrects the auto-collisions for the outer surface(s) of a 3D epithelium.

    Parameters
    ----------
    eptm : a :class:`Epithelium` object
    position_buffer : np.array of shape (eptm.Nv, eptm.dim):
        positions of the vertices prior to the collisions

    Returns
    -------
    changed : bool
        `True` if the positions of some vertices were changed

    """

    sub_sheet = get_outer_sheet(eptm)
    pos_idx = sub_sheet.vert_df.index.copy()
    sub_sheet.reset_index()
    sub_buffer = pd.DataFrame(
        position_buffer.loc[pos_idx].values,
        index=sub_sheet.vert_df.index,
        columns=sub_sheet.coords,
    )
    changed = solve_sheet_collisions(sub_sheet, sub_buffer)
    if changed:
        eptm.vert_df.loc[pos_idx, eptm.coords] = sub_sheet.vert_df[eptm.coords].values
    return changed


def solve_sheet_collisions(sheet, position_buffer):
    """Corrects the auto-collisions for the outer surface(s) of a 2.5D sheet.

    Parameters
    ----------
    sheet : a :class:`Sheet` object
    position_buffer : np.array of shape (sheet.Nv, sheet.dim):
        positions of the vertices prior to the collisions

    Returns
    -------
    changed : bool
        `True` if the positions of some vertices were changed

    """

    intersecting_edges = self_intersections(sheet)
    if intersecting_edges.shape[0]:
        log.info("%d intersections were detected", intersecting_edges.shape[0])
        shyness = sheet.settings.get("shyness", 1e-10)
        boxes = CollidingBoxes(sheet, position_buffer, intersecting_edges)
        changed = boxes.solve_collisions(shyness)
        return changed
    return False


class CollidingBoxes:
    """Utility class to manage collisions
    """

    def __init__(self, sheet, position_buffer, intersecting_edges):
        """Creates a CollidingBoxes instance

        Parameters
        ----------

        sheet : a :clas:`Sheet` instance
        position_buffer : np.array of shape (sheet.Nv, sheet.dim):
            positions of the vertices prior to the collisions
        intersecting_edges : np.ndarray
            pairs of indices of the intersecting edges

        """
        self.sheet = sheet
        self.edge_pairs = intersecting_edges
        self.face_pairs = self._get_intersecting_faces()
        self.edge_buffer = sheet.upcast_srce(position_buffer).copy()
        self.edge_buffer.columns = ["sx", "sy", "sz"]
        self.plane_not_found = False

    def _get_intersecting_faces(self):
        """Returns unique pairs of intersecting faces

        """
        _face_pairs = self.sheet.edge_df.loc[
            self.edge_pairs.flatten(), "face"
        ].values.reshape((-1, 2))
        unique_pairs = set(map(frozenset, _face_pairs))

        return np.array([[*pair] for pair in unique_pairs if len(pair) == 2])

    def get_limits(self, shyness=1e-10):
        """ Iterator over the position boundaries avoiding the
        collisions.

        Parameters
        ----------
        shyness : float
          the extra distance between two colliding vertices,
          on each side of the collision plane.

        Yields
        ------
        lower, upper : two `pd.Series`
           those Series are indexed by the vertices of the colliding
           faces giving the lower and upper bounds for the vertices
        """
        for face_pair in self.face_pairs:
            yield self._collision_plane(face_pair, shyness)

    def solve_collisions(self, shyness=1e-10):
        """ Solves the collisions by finding the collision plane.

        Modifies the sheet vertex positions inplace such that they
        rest at a distance ``shyness`` apart on each side of the collision plane.

        Parameters
        ----------
        shyness : float, default 1e-10
          the extra distance between two colliding vertices,
          on each side of the collision plane.

        Based on Liu, J.-D., Ko, M.-T., & Chang, R.-C. (1998),
        *A simple self-collision avoidance for cloth animation*.
        Computers & Graphics, 22(1), 117–128.
        `DOI <https://doi.org/doi:10.1016/s0097-8493(97)00087-3>`_

        """

        colliding_verts = self.sheet.edge_df[
            self.sheet.edge_df["face"].isin(self.face_pairs.ravel())
        ]["srce"].unique()
        upper_bounds = pd.DataFrame(
            index=pd.Index(colliding_verts, name="vert"), columns=self.sheet.coords
        )
        upper_bounds[:] = np.inf
        lower_bounds = pd.DataFrame(
            index=pd.Index(colliding_verts, name="vert"), columns=self.sheet.coords
        )
        lower_bounds[:] = -np.inf
        plane_found = False
        for lower, upper in self.get_limits(shyness):
            if lower is None:
                continue
            plane_found = True
            sub_lower = lower_bounds.loc[lower.index, lower.columns]
            lower_bounds.loc[lower.index, lower.columns] = np.maximum(sub_lower, lower)

            sub_upper = upper_bounds.loc[upper.index, upper.columns]
            upper_bounds.loc[upper.index, upper.columns] = np.minimum(sub_upper, upper)

        if upper_bounds.shape[0] == 0:
            plane_found = False

        if not plane_found:
            return False

        upper_bounds.index.name = "vert"
        upper_bounds = (
            upper_bounds[np.isfinite(upper_bounds.values.astype(float))]
            .groupby("vert")
            .apply(min)
        )
        lower_bounds.index.name = "vert"
        lower_bounds = (
            lower_bounds[np.isfinite(lower_bounds.values.astype(float))]
            .groupby("vert")
            .apply(max)
        )

        correction_upper = np.minimum(
            self.sheet.vert_df.loc[upper_bounds.index, self.sheet.coords],
            upper_bounds.values,
        )
        correction_lower = np.maximum(
            self.sheet.vert_df.loc[lower_bounds.index, self.sheet.coords],
            lower_bounds.values,
        )
        corrections = pd.concat((correction_lower, correction_upper), axis=0)
        self.sheet.vert_df.loc[
            corrections.index.values, self.sheet.coords
        ] = corrections
        return True

    def _collision_plane(self, face_pair, shyness):

        f0, f1 = face_pair

        fe0c = self.sheet.edge_df[self.sheet.edge_df["face"] == f0].copy()
        fe1c = self.sheet.edge_df[self.sheet.edge_df["face"] == f1].copy()
        fe0p = self.edge_buffer[self.sheet.edge_df["face"] == f0].copy()
        fe1p = self.edge_buffer[self.sheet.edge_df["face"] == f1].copy()

        bb0c = _face_bbox(fe0c)
        bb1c = _face_bbox(fe1c)
        bb0p = _face_bbox(fe0p)
        bb1p = _face_bbox(fe1p)

        dr0 = bb0c - bb0p
        dr1 = bb1c - bb1p
        sign_change_l1h0 = np.sign((bb1c.l - bb0c.h) * (bb1p.l - bb0p.h)) < 0
        sign_change_l0h1 = np.sign((bb0c.l - bb1c.h) * (bb0p.l - bb1p.h)) < 0

        # face 0 is to the left of face 1 on the collision axis
        if any(sign_change_l1h0):
            lower_bound = pd.DataFrame(index=fe1c.srce)
            upper_bound = pd.DataFrame(index=fe0c.srce)

            coll_ax = bb0c[sign_change_l1h0].index
            plane_coords = ((bb0p.h * dr1.l - bb1p.l * dr0.h) / (dr1.l - dr0.h)).loc[
                coll_ax
            ]
        # face 0 is to the right of face 1 on the collision axis
        elif any(sign_change_l0h1):
            lower_bound = pd.DataFrame(index=fe0c.srce)
            upper_bound = pd.DataFrame(index=fe1c.srce)
            coll_ax = bb0c[sign_change_l0h1].index
            plane_coords = ((bb1p.h * dr0.l - bb0p.l * dr1.h) / (dr0.l - dr1.h)).loc[
                coll_ax
            ]
        else:
            log.info("""Plane Not Found""")
            self.plane_not_found = True
            lower_bound = pd.DataFrame(
                index=set(fe0c.srce).union(fe1c.srce), columns=list("xyz")
            )
            upper_bound = pd.DataFrame(
                index=set(fe0c.srce).union(fe1c.srce), columns=list("xyz")
            )
            for c in list("xyz"):
                b0 = bb0c.loc[c]
                b1 = bb1c.loc[c]
                left, right = (fe0c, fe1c) if (b0.mean() < b1.mean()) else (fe1c, fe0c)

                lim = (left[f"s{c}"].max() + right[f"s{c}"].min()) / 2
                upper_bound.loc[right.srce, c] = right[f"s{c}"].max()
                upper_bound.loc[left.srce, c] = lim - shyness / 2

                lower_bound.loc[left.srce, c] = left[f"s{c}"].min()
                lower_bound.loc[right.srce, c] = lim + shyness / 2

            return lower_bound, upper_bound

        for c in coll_ax:
            lower_bound[c] = plane_coords.loc[c] + shyness / 2
            upper_bound[c] = plane_coords.loc[c] - shyness / 2

        return lower_bound, upper_bound


def _face_bbox(face_edges):

    points = face_edges[["sx", "sy", "sz"]].values
    lower = points.min(axis=0)
    upper = points.max(axis=0)
    return pd.DataFrame(
        [lower, upper], index=list("lh"), columns=list("xyz"), dtype=np.float
    ).T


def revert_positions(sheet, position_buffer, intersecting_edges):

    unique_edges = np.unique(intersecting_edges)
    unique_verts = np.unique(sheet.edge_df.loc[unique_edges, ["srce", "trgt"]])
    sheet.vert_df.loc[unique_verts, sheet.coords] = position_buffer.loc[unique_verts]
